from void_terminal.toolbox import CatchException, update_ui
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive, input_clipping
import requests
from bs4 import BeautifulSoup
from void_terminal.request_llms.bridge_all import model_info

def google(query, proxies):
    query = query # Replace the keywords you want to search here
    url = f"https://www.google.com/search?q={query}"
    headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.61 Safari/537.36'}
    response = requests.get(url, headers=headers, proxies=proxies)
    soup = BeautifulSoup(response.content, 'html.parser')
    results = []
    for g in soup.find_all('div', class_='g'):
        anchors = g.find_all('a')
        if anchors:
            link = anchors[0]['href']
            if link.startswith('/url?q='):
                link = link[7:]
            if not link.startswith('http'):
                continue
            title = g.find('h3').text
            item = {'title': title, 'link': link}
            results.append(item)

    for r in results:
        print(r['link'])
    return results

def scrape_text(url, proxies) -> str:
    """Scrape text from a webpage

    Args:
        url (str): The URL to scrape text from

    Returns:
        str: The scraped text
    """
    headers = {
        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.61 Safari/537.36',
        'Content-Type': 'text/plain',
    }
    try: 
        response = requests.get(url, headers=headers, proxies=proxies, timeout=8)
        if response.encoding == "ISO-8859-1": response.encoding = response.apparent_encoding
    except: 
        return "Cannot connect to the webpage"
    soup = BeautifulSoup(response.text, "html.parser")
    for script in soup(["script", "style"]):
        script.extract()
    text = soup.get_text()
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    text = "\n".join(chunk for chunk in chunks if chunk)
    return text

@CatchException
def ConnectToNetworkToAnswerQuestions(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    """
    txt             Text entered by the user in the input field，For example, a paragraph that needs to be translated，For example, a file path that contains files to be processed
    llm_kwargs      GPT model parameters，Such as temperature and top_p，Generally pass it on as is
    plugin_kwargs   Plugin model parameters，No use for the time being
    chatbot         Chat display box handle，Displayed to the user
    history         Chat history，Context summary
    system_prompt   Silent reminder to GPT
    web_port        Current software running port number
    """
    history = []    # Clear history，To avoid input overflow
    chatbot.append((f"Please answer the following questions based on internet information：{txt}", 
                    "[Local Message] Please note，You are calling a[function plugin]template，This template can achieve ChatGPT network information integration。This function is aimed at developers who want to implement more interesting features，It can be used as a template for creating new feature functions。If you want to share new feature modules，Please don`t hesitate to PR!"))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page # As requesting GPT takes some time，Let`s do a UI update in time

    # ------------- < Step 1：Crawl the results of search engines > -------------
    from void_terminal.toolbox import get_conf
    proxies = get_conf('proxies')
    urls = google(txt, proxies)
    history = []
    if len(urls) == 0:
        chatbot.append((f"Conclusion：{txt}",
                        "[Local Message] Restricted by Google，Unable to retrieve information from Google!"))
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page # As requesting GPT takes some time，Let`s do a UI update in time
        return
    # ------------- < Step 2：Visit web pages in order > -------------
    max_search_result = 5   # Include results from how many web pages at most
    for index, url in enumerate(urls[:max_search_result]):
        res = scrape_text(url['link'], proxies)
        history.extend([f"The{index}search results：", res])
        chatbot.append([f"The{index}search results：", res[:500]+"......"])
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page # As requesting GPT takes some time，Let`s do a UI update in time

    # ------------- < Step 3：ChatGPT synthesis > -------------
    i_say = f"Extract information from the above search results，Then answer the question：{txt}"
    i_say, history = input_clipping(    # Trim the input，Start trimming from the longest entry，Prevent token explosion
        inputs=i_say, 
        history=history, 
        max_token_limit=model_info[llm_kwargs['llm_model']]['max_token']*3//4
    )
    gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
        inputs=i_say, inputs_show_user=i_say, 
        llm_kwargs=llm_kwargs, chatbot=chatbot, history=history, 
        sys_prompt="Please extract information from the given search results，Summarize the two most relevant search results，Then answer the question。"
    )
    chatbot[-1] = (i_say, gpt_say)
    history.append(i_say);history.append(gpt_say)
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page # UI update




from void_terminal.toolbox import get_conf
import base64
import datetime
import hashlib
import hmac
import json
from urllib.parse import urlparse
import ssl
from datetime import datetime
from time import mktime
from urllib.parse import urlencode
from wsgiref.handlers import format_date_time
import websocket
import threading, time

timeout_bot_msg = '[Local Message] Request timeout. Network error.'

class Ws_Param(object):
    # Initialization
    def __init__(self, APPID, APIKey, APISecret, gpt_url):
        self.APPID = APPID
        self.APIKey = APIKey
        self.APISecret = APISecret
        self.host = urlparse(gpt_url).netloc
        self.path = urlparse(gpt_url).path
        self.gpt_url = gpt_url

    # Generate URL
    def create_url(self):
        # Generate RFC1123 formatted timestamp
        now = datetime.now()
        date = format_date_time(mktime(now.timetuple()))

        # Concatenate strings
        signature_origin = "host: " + self.host + "\n"
        signature_origin += "date: " + date + "\n"
        signature_origin += "GET " + self.path + " HTTP/1.1"

        # Perform encryption using hmac-sha256
        signature_sha = hmac.new(self.APISecret.encode('utf-8'), signature_origin.encode('utf-8'), digestmod=hashlib.sha256).digest()
        signature_sha_base64 = base64.b64encode(signature_sha).decode(encoding='utf-8')
        authorization_origin = f'api_key="{self.APIKey}", algorithm="hmac-sha256", headers="host date request-line", signature="{signature_sha_base64}"'
        authorization = base64.b64encode(authorization_origin.encode('utf-8')).decode(encoding='utf-8')

        # Combine the requested authentication parameters into a dictionary
        v = {
            "authorization": authorization,
            "date": date,
            "host": self.host
        }
        # Concatenate authentication parameters，Generate URL
        url = self.gpt_url + '?' + urlencode(v)
        # Print the URL when establishing the connection here,You can remove the comments above when referring to this demo，Check if the generated URL matches the one generated by your code when comparing the same parameters
        return url



class SparkRequestInstance():
    def __init__(self):
        XFYUN_APPID, XFYUN_API_SECRET, XFYUN_API_KEY = get_conf('XFYUN_APPID', 'XFYUN_API_SECRET', 'XFYUN_API_KEY')
        if XFYUN_APPID == '00000000' or XFYUN_APPID == '': raise RuntimeError('Please configure XFYUN_APPID for the Xunfei Starfire model, XFYUN_API_KEY, XFYUN_API_SECRET')
        self.appid = XFYUN_APPID
        self.api_secret = XFYUN_API_SECRET
        self.api_key = XFYUN_API_KEY
        self.gpt_url = "ws://spark-api.xf-yun.com/v1.1/chat"
        self.gpt_url_v2 = "ws://spark-api.xf-yun.com/v2.1/chat"
        self.gpt_url_v3 = "ws://spark-api.xf-yun.com/v3.1/chat"

        self.time_to_yield_event = threading.Event()
        self.time_to_exit_event = threading.Event()

        self.result_buf = ""

    def generate(self, inputs, llm_kwargs, history, system_prompt):
        llm_kwargs = llm_kwargs
        history = history
        system_prompt = system_prompt
        import _thread as thread
        thread.start_new_thread(self.create_blocking_request, (inputs, llm_kwargs, history, system_prompt))
        while True:
            self.time_to_yield_event.wait(timeout=1)
            if self.time_to_yield_event.is_set():
                yield self.result_buf
            if self.time_to_exit_event.is_set():
                return self.result_buf


    def create_blocking_request(self, inputs, llm_kwargs, history, system_prompt):
        if llm_kwargs['llm_model'] == 'sparkv2':
            gpt_url = self.gpt_url_v2
        elif llm_kwargs['llm_model'] == 'sparkv3':
            gpt_url = self.gpt_url_v3
        else:
            gpt_url = self.gpt_url

        wsParam = Ws_Param(self.appid, self.api_key, self.api_secret, gpt_url)
        websocket.enableTrace(False)
        wsUrl = wsParam.create_url()

        # Processing when receiving websocket connection establishment
        def on_open(ws):
            import _thread as thread
            thread.start_new_thread(run, (ws,))

        def run(ws, *args):
            data = json.dumps(gen_params(ws.appid, *ws.all_args))
            ws.send(data)

        # Processing of received websocket messages
        def on_message(ws, message):
            data = json.loads(message)
            code = data['header']['code']
            if code != 0:
                print(f'Request error: {code}, {data}')
                self.result_buf += str(data)
                ws.close()
                self.time_to_exit_event.set()
            else:
                choices = data["payload"]["choices"]
                status = choices["status"]
                content = choices["text"][0]["content"]
                ws.content += content
                self.result_buf += content
                if status == 2:
                    ws.close()
                    self.time_to_exit_event.set()
            self.time_to_yield_event.set()

        # Handling websocket errors
        def on_error(ws, error):
            print("error:", error)
            self.time_to_exit_event.set()

        # Processing when receiving websocket closure
        def on_close(ws, *args):
            self.time_to_exit_event.set()

        # websocket
        ws = websocket.WebSocketApp(wsUrl, on_message=on_message, on_error=on_error, on_close=on_close, on_open=on_open)
        ws.appid = self.appid
        ws.content = ""
        ws.all_args = (inputs, llm_kwargs, history, system_prompt)
        ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})

def generate_message_payload(inputs, llm_kwargs, history, system_prompt):
    conversation_cnt = len(history) // 2
    messages = [{"role": "system", "content": system_prompt}]
    if conversation_cnt:
        for index in range(0, 2*conversation_cnt, 2):
            what_i_have_asked = {}
            what_i_have_asked["role"] = "user"
            what_i_have_asked["content"] = history[index]
            what_gpt_answer = {}
            what_gpt_answer["role"] = "assistant"
            what_gpt_answer["content"] = history[index+1]
            if what_i_have_asked["content"] != "":
                if what_gpt_answer["content"] == "": continue
                if what_gpt_answer["content"] == timeout_bot_msg: continue
                messages.append(what_i_have_asked)
                messages.append(what_gpt_answer)
            else:
                messages[-1]['content'] = what_gpt_answer['content']
    what_i_ask_now = {}
    what_i_ask_now["role"] = "user"
    what_i_ask_now["content"] = inputs
    messages.append(what_i_ask_now)
    return messages


def gen_params(appid, inputs, llm_kwargs, history, system_prompt):
    """
    Generate request parameters through appid and user`s question
    """
    domains = {
        "spark": "general",
        "sparkv2": "generalv2",
        "sparkv3": "generalv3",
    }
    data = {
        "header": {
            "app_id": appid,
            "uid": "1234"
        },
        "parameter": {
            "chat": {
                "domain": domains[llm_kwargs['llm_model']],
                "temperature": llm_kwargs["temperature"],
                "random_threshold": 0.5,
                "max_tokens": 4096,
                "auditing": "default"
            }
        },
        "payload": {
            "message": {
                "text": generate_message_payload(inputs, llm_kwargs, history, system_prompt)
            }
        }
    }
    return data




model_name = "InternLM"
cmd_to_install = "`pip install -r request_llms/requirements_chatglm.txt`"

from transformers import AutoModel, AutoTokenizer
import time
import threading
import importlib
from void_terminal.toolbox import update_ui, get_conf, ProxyNetworkActivate
from multiprocessing import Process, Pipe
from void_terminal.request_llms.local_llm_class import LocalLLMHandle, get_local_llm_predict_fns


# ------------------------------------------------------------------------------------------------------------------------
# 🔌💻 Local Model Utils
# ------------------------------------------------------------------------------------------------------------------------
def try_to_import_special_deps():
    import sentencepiece

def combine_history(prompt, hist):
    user_prompt = "<|User|>:{user}<eoh>\n"
    robot_prompt = "<|Bot|>:{robot}<eoa>\n"
    cur_query_prompt = "<|User|>:{user}<eoh>\n<|Bot|>:"
    messages = hist
    total_prompt = ""
    for message in messages:
        cur_content = message
        cur_prompt = user_prompt.replace("{user}", cur_content[0])
        total_prompt += cur_prompt
        cur_prompt = robot_prompt.replace("{robot}", cur_content[1])
        total_prompt += cur_prompt
    total_prompt = total_prompt + cur_query_prompt.replace("{user}", prompt)
    return total_prompt

# ------------------------------------------------------------------------------------------------------------------------
# 🔌💻 Local Model
# ------------------------------------------------------------------------------------------------------------------------
class GetInternlmHandle(LocalLLMHandle):

    def load_model_info(self):
        # 🏃‍♂️🏃‍♂️🏃‍♂️ Subprocess execution
        self.model_name = model_name
        self.cmd_to_install = cmd_to_install

    def try_to_import_special_deps(self, **kwargs):
        """
        import something that will raise error if the user does not install requirement_*.txt
        """
        import sentencepiece

    def load_model_and_tokenizer(self):
        # 🏃‍♂️🏃‍♂️🏃‍♂️ Subprocess execution
        import torch
        from transformers import AutoModelForCausalLM, AutoTokenizer
        device = get_conf('LOCAL_MODEL_DEVICE')
        with ProxyNetworkActivate('Download_LLM'):
            if self._model is None:
                tokenizer = AutoTokenizer.from_pretrained("internlm/internlm-chat-7b", trust_remote_code=True)
                if device=='cpu':
                    model = AutoModelForCausalLM.from_pretrained("internlm/internlm-chat-7b", trust_remote_code=True).to(torch.bfloat16)
                else:
                    model = AutoModelForCausalLM.from_pretrained("internlm/internlm-chat-7b", trust_remote_code=True).to(torch.bfloat16).cuda()

                model = model.eval()
        return model, tokenizer

    def llm_stream_generator(self, **kwargs):
        import torch
        import logging
        import copy
        import warnings
        import torch.nn as nn
        from transformers.generation.utils import LogitsProcessorList, StoppingCriteriaList, GenerationConfig

        # 🏃‍♂️🏃‍♂️🏃‍♂️ Subprocess execution
        def adaptor():
            model = self._model
            tokenizer = self._tokenizer
            prompt = kwargs['query']
            max_length = kwargs['max_length']
            top_p = kwargs['top_p']
            temperature = kwargs['temperature']
            history = kwargs['history']
            real_prompt = combine_history(prompt, history)
            return model, tokenizer, real_prompt, max_length, top_p, temperature
        
        model, tokenizer, prompt, max_length, top_p, temperature = adaptor()
        prefix_allowed_tokens_fn = None
        logits_processor = None
        stopping_criteria = None
        additional_eos_token_id = 103028
        generation_config = None
        # 🏃‍♂️🏃‍♂️🏃‍♂️ Subprocess execution
        # 🏃‍♂️🏃‍♂️🏃‍♂️ https://github.com/InternLM/InternLM/blob/efbf5335709a8c8faeac6eaf07193973ff1d56a1/web_demo.py#L25

        inputs = tokenizer([prompt], padding=True, return_tensors="pt")
        input_length = len(inputs["input_ids"][0])
        device = get_conf('LOCAL_MODEL_DEVICE')
        for k, v in inputs.items():
            inputs[k] = v.to(device)
        input_ids = inputs["input_ids"]
        batch_size, input_ids_seq_length = input_ids.shape[0], input_ids.shape[-1]
        if generation_config is None:
            generation_config = model.generation_config
        generation_config = copy.deepcopy(generation_config)
        model_kwargs = generation_config.update(**kwargs)
        bos_token_id, eos_token_id = generation_config.bos_token_id, generation_config.eos_token_id
        if isinstance(eos_token_id, int):
            eos_token_id = [eos_token_id]
        if additional_eos_token_id is not None:
            eos_token_id.append(additional_eos_token_id)
        has_default_max_length = kwargs.get("max_length") is None and generation_config.max_length is not None
        if has_default_max_length and generation_config.max_new_tokens is None:
            warnings.warn(
                f"Using `max_length`'s default ({generation_config.max_length}) to control the generation length. "
                "This behaviour is deprecated and will be removed from the config in v5 of Transformers -- we"
                " recommend using `max_new_tokens` to control the maximum length of the generation.",
                UserWarning,
            )
        elif generation_config.max_new_tokens is not None:
            generation_config.max_length = generation_config.max_new_tokens + input_ids_seq_length
            if not has_default_max_length:
                logging.warn(
                    f"Both `max_new_tokens` (={generation_config.max_new_tokens}) and `max_length`(="
                    f"{generation_config.max_length}) seem to have been set. `max_new_tokens` will take precedence. "
                    "Please refer to the documentation for more information. "
                    "(https://huggingface.co/docs/transformers/main/en/main_classes/text_generation)",
                    UserWarning,
                )

        if input_ids_seq_length >= generation_config.max_length:
            input_ids_string = "input_ids"
            logging.warning(
                f"Input length of {input_ids_string} is {input_ids_seq_length}, but `max_length` is set to"
                f" {generation_config.max_length}. This can lead to unexpected behavior. You should consider"
                " increasing `max_new_tokens`."
            )

        # 2. Set generation parameters if not already defined
        logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
        stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()

        logits_processor = model._get_logits_processor(
            generation_config=generation_config,
            input_ids_seq_length=input_ids_seq_length,
            encoder_input_ids=input_ids,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            logits_processor=logits_processor,
        )

        stopping_criteria = model._get_stopping_criteria(
            generation_config=generation_config, stopping_criteria=stopping_criteria
        )
        logits_warper = model._get_logits_warper(generation_config)

        unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)
        scores = None
        while True:
            model_inputs = model.prepare_inputs_for_generation(input_ids, **model_kwargs)
            # forward pass to get next token
            outputs = model(
                **model_inputs,
                return_dict=True,
                output_attentions=False,
                output_hidden_states=False,
            )

            next_token_logits = outputs.logits[:, -1, :]

            # pre-process distribution
            next_token_scores = logits_processor(input_ids, next_token_logits)
            next_token_scores = logits_warper(input_ids, next_token_scores)

            # sample
            probs = nn.functional.softmax(next_token_scores, dim=-1)
            if generation_config.do_sample:
                next_tokens = torch.multinomial(probs, num_samples=1).squeeze(1)
            else:
                next_tokens = torch.argmax(probs, dim=-1)

            # update generated ids, model inputs, and length for next step
            input_ids = torch.cat([input_ids, next_tokens[:, None]], dim=-1)
            model_kwargs = model._update_model_kwargs_for_generation(
                outputs, model_kwargs, is_encoder_decoder=False
            )
            unfinished_sequences = unfinished_sequences.mul((min(next_tokens != i for i in eos_token_id)).long())
            
            output_token_ids = input_ids[0].cpu().tolist()
            output_token_ids = output_token_ids[input_length:]
            for each_eos_token_id in eos_token_id:
                if output_token_ids[-1] == each_eos_token_id:
                    output_token_ids = output_token_ids[:-1]
            response = tokenizer.decode(output_token_ids)

            yield response
            # stop when each sentence is finished, or if we exceed the maximum length
            if unfinished_sequences.max() == 0 or stopping_criteria(input_ids, scores):
                return

    
# ------------------------------------------------------------------------------------------------------------------------
# 🔌💻 GPT-Academic Interface
# ------------------------------------------------------------------------------------------------------------------------
predict_no_ui_long_connection, predict = get_local_llm_predict_fns(GetInternlmHandle, model_name)


"""
Explanation of the Void Terminal Plugin:

Please describe in natural language what you want to do.

1. You can open the plugin's dropdown menu to explore various capabilities of this project, and then describe your needs in natural language, for example:
- "Please call the plugin to translate a PDF paper for me. I just uploaded the paper to the upload area."
- "Please use the plugin to translate a PDF paper, with the address being https://www.nature.com/articles/s41586-019-1724-z.pdf."
- "Generate an image with blooming flowers and lush green grass using the plugin."
- "Translate the README using the plugin. The GitHub URL is https://github.com/facebookresearch/co-tracker."
- "Translate an Arxiv paper for me. The Arxiv ID is 1812.10695. Remember to use the plugin and don't do it manually!"
- "I don't like the current interface color. Modify the configuration and change the theme to THEME="High-Contrast"."
- "Could you please explain the structure of the Transformer network?"

2. If you use keywords like "call the plugin xxx", "modify the configuration xxx", "please", etc., your intention can be recognized more accurately.

3. Your intention can be recognized more accurately when using powerful models like GPT4. This plugin is relatively new, so please feel free to provide feedback on GitHub.

4. Now, if you need to process a file, please upload the file (drag the file to the file upload area) or describe the path to the file.

5. If you don't need to upload a file, you can simply repeat your command again.
"""
explain_msg = """
## VoidTerminal plugin description:

1. Please describe what you need to do in **natural language**。For example：
    - Please call the plugin，Translate PDF papers for me，I just put the paper in the upload area
    - Please call the plugin to translate the PDF paper，The address is https://openreview.net/pdf?id=rJl0r3R9KX」
    - Translate Arxiv papers into Chinese PDF，The ID of the arxiv paper is 1812.10695，Remember to use the plugin!
    - Generate an image，Flowers blooming in the picture，Green grass，Implemented with a plugin
    - Translate README with plugins，The Github URL is https://github.com/facebookresearch/co-tracker」
    - I don`t like the current interface color，Modify configuration，把主题THEME更换为THEME="High-Contrast"」
    - Please call the plugin，Parsing Python source code project，I just packed the code and dragged it to the upload area
    - What is the structure of the Transformer network?？」

2. You can open the plugin dropdown menu to learn about various capabilities of this project。    

3. If you use keywords such as `call plugin xxx`, `modify configuration xxx`, `please`, etc.，Your intent can be recognized more accurately。

4. It is recommended to use GPT3.5 or a stronger model，Weak model may not understand your ideas。This plugin has not been around for long，Welcome to go to Github to provide feedback。

5. Now，If file processing is required，Please upload the file（Drag and drop the file to the file upload area），Or the path of the description file。

6. If file upload is not needed，Now you just need to repeat your command again。
"""

from pydantic import BaseModel, Field
from typing import List
from void_terminal.toolbox import CatchException, update_ui, is_the_upload_folder
from void_terminal.toolbox import update_ui_lastest_msg, disable_auto_promotion
from void_terminal.request_llms.bridge_all import predict_no_ui_long_connection
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive
from void_terminal.crazy_functions.crazy_utils import input_clipping
from void_terminal.crazy_functions.json_fns.pydantic_io import GptJsonIO, JsonStringError
from void_terminal.crazy_functions.vt_fns.vt_state import VoidTerminalState
from void_terminal.crazy_functions.vt_fns.vt_modify_config import modify_configuration_hot
from void_terminal.crazy_functions.vt_fns.vt_modify_config import modify_configuration_reboot
from void_terminal.crazy_functions.vt_fns.vt_call_plugin import execute_plugin

class UserIntention(BaseModel):
    user_prompt: str = Field(description="the content of user input", default="")
    intention_type: str = Field(description="the type of user intention, choose from ['ModifyConfiguration', 'ExecutePlugin', 'Chat']", default="ExecutePlugin")
    user_provide_file: bool = Field(description="whether the user provides a path to a file", default=False)
    user_provide_url: bool = Field(description="whether the user provides a url", default=False)


def chat(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, user_intention):
    gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
        inputs=txt, inputs_show_user=txt,
        llm_kwargs=llm_kwargs, chatbot=chatbot, history=[], 
        sys_prompt=system_prompt
    )
    chatbot[-1] = [txt, gpt_say]
    history.extend([txt, gpt_say])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
    pass


explain_intention_to_user = {
    'Chat': "Chat conversation",
    'ExecutePlugin': "Invoke plugin",
    'ModifyConfiguration': "Modify configuration",
}


def analyze_intention_with_simple_rules(txt):
    user_intention = UserIntention()
    user_intention.user_prompt = txt
    is_certain = False

    if 'May I ask' in txt:
        is_certain = True
        user_intention.intention_type = 'Chat'

    if 'Use Plugin' in txt:
        is_certain = True
        user_intention.intention_type = 'ExecutePlugin'

    if 'Modify configuration' in txt:
        is_certain = True
        user_intention.intention_type = 'ModifyConfiguration'

    return is_certain, user_intention


@CatchException
def VoidTerminal(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    disable_auto_promotion(chatbot=chatbot)
    # Get the current VoidTerminal status
    state = VoidTerminalState.get_state(chatbot)
    appendix_msg = ""

    # Detect user intention with simple keywords
    is_certain, _ = analyze_intention_with_simple_rules(txt)
    if is_the_upload_folder(txt):
        state.set_state(chatbot=chatbot, key='has_provided_explaination', value=False)
        appendix_msg = "\n\n**Very good，You have uploaded the file **，Now please describe your requirements。"
        
    if is_certain or (state.has_provided_explaination):
        # If the intention is clear，Skip prompt section
        state.set_state(chatbot=chatbot, key='has_provided_explaination', value=True)
        state.unlock_plugin(chatbot=chatbot)
        yield from update_ui(chatbot=chatbot, history=history)
        yield from VoidTerminalMainRoute(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port)
        return
    else:
        # If the intent is ambiguous，prompt
        state.set_state(chatbot=chatbot, key='has_provided_explaination', value=True)
        state.lock_plugin(chatbot=chatbot)
        chatbot.append(("VoidTerminal status:", explain_msg+appendix_msg))
        yield from update_ui(chatbot=chatbot, history=history)
        return



def VoidTerminalMainRoute(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    history = []
    chatbot.append(("VoidTerminal status: ", f"Executing task: {txt}"))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # ⭐ ⭐ ⭐ Analyze user intent
    is_certain, user_intention = analyze_intention_with_simple_rules(txt)
    if not is_certain:
        yield from update_ui_lastest_msg(
            lastmsg=f"Executing task: {txt}\n\nAnalyzing user intent", chatbot=chatbot, history=history, delay=0)
        gpt_json_io = GptJsonIO(UserIntention)
        rf_req = "\nchoose from ['ModifyConfiguration', 'ExecutePlugin', 'Chat']"
        inputs = "Analyze the intention of the user according to following user input: \n\n" + \
            ">> " + (txt+rf_req).rstrip('\n').replace('\n','\n>> ') + '\n\n' + gpt_json_io.format_instructions
        run_gpt_fn = lambda inputs, sys_prompt: predict_no_ui_long_connection(
            inputs=inputs, llm_kwargs=llm_kwargs, history=[], sys_prompt=sys_prompt, observe_window=[])
        analyze_res = run_gpt_fn(inputs, "")
        try:
            user_intention = gpt_json_io.generate_output_auto_repair(analyze_res, run_gpt_fn)
            lastmsg=f"Executing task: {txt}\n\nUser intent understanding: Intent ={explain_intention_to_user[user_intention.intention_type]}", 
        except JsonStringError as e:
            yield from update_ui_lastest_msg(
                lastmsg=f"Executing task: {txt}\n\nUser intent understanding: Failed current language model（{llm_kwargs['llm_model']}）Cannot understand your intention", chatbot=chatbot, history=history, delay=0)
            return
    else:
        pass

    yield from update_ui_lastest_msg(
        lastmsg=f"Executing task: {txt}\n\nUser intent understanding: Intent ={explain_intention_to_user[user_intention.intention_type]}", 
        chatbot=chatbot, history=history, delay=0)

    # User intent: Modify the configuration of this project
    if user_intention.intention_type == 'ModifyConfiguration':
        yield from modify_configuration_reboot(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, user_intention)

    # User intent: Scheduling plugin
    if user_intention.intention_type == 'ExecutePlugin':
        yield from execute_plugin(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, user_intention)

    # User intent: Chat
    if user_intention.intention_type == 'Chat':
        yield from chat(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, user_intention)

    return




from void_terminal.toolbox import get_log_folder, update_ui, gen_time_str, get_conf, promote_file_to_downloadzone
from void_terminal.crazy_functions.agent_fns.watchdog import WatchDog
import time, os

class PipeCom:
    def __init__(self, cmd, content) -> None:
        self.cmd = cmd
        self.content = content


class PluginMultiprocessManager:
    def __init__(self, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
        # ⭐ run in main process
        self.autogen_work_dir = os.path.join(get_log_folder("autogen"), gen_time_str())
        self.previous_work_dir_files = {}
        self.llm_kwargs = llm_kwargs
        self.plugin_kwargs = plugin_kwargs
        self.chatbot = chatbot
        self.history = history
        self.system_prompt = system_prompt
        # self.web_port = web_port
        self.alive = True
        self.use_docker = get_conf("AUTOGEN_USE_DOCKER")
        self.last_user_input = ""
        # create a thread to monitor self.heartbeat, terminate the instance if no heartbeat for a long time
        timeout_seconds = 5 * 60
        self.heartbeat_watchdog = WatchDog(timeout=timeout_seconds, bark_fn=self.terminate, interval=5)
        self.heartbeat_watchdog.begin_watch()

    def feed_heartbeat_watchdog(self):
        # feed this `dog`, so the dog will not `bark` (bark_fn will terminate the instance)
        self.heartbeat_watchdog.feed()

    def is_alive(self):
        return self.alive

    def launch_subprocess_with_pipe(self):
        # ⭐ run in main process
        from multiprocessing import Process, Pipe

        parent_conn, child_conn = Pipe()
        self.p = Process(target=self.subprocess_worker, args=(child_conn,))
        self.p.daemon = True
        self.p.start()
        return parent_conn

    def terminate(self):
        self.p.terminate()
        self.alive = False
        print("[debug] instance terminated")

    def subprocess_worker(self, child_conn):
        # ⭐⭐ run in subprocess
        raise NotImplementedError

    def send_command(self, cmd):
        # ⭐ run in main process
        repeated = False
        if cmd == self.last_user_input:
            repeated = True
            cmd = ""
        else:
            self.last_user_input = cmd
        self.parent_conn.send(PipeCom("user_input", cmd))
        return repeated, cmd

    def immediate_showoff_when_possible(self, fp):
        # ⭐ Main process
        # Get the extension name of fp
        file_type = fp.split('.')[-1]
        # If it is a text file, Then display the text content directly
        if file_type.lower() in ['png', 'jpg']:
            image_path = os.path.abspath(fp)
            self.chatbot.append([
                'Detected new image:', 
                f'Local file preview: <br/><div align="center"><img src="file={image_path}"></div>'
            ])
            yield from update_ui(chatbot=self.chatbot, history=self.history)

    def overwatch_workdir_file_change(self):
        # ⭐ Main Process Docker External Folder Monitoring
        path_to_overwatch = self.autogen_work_dir
        change_list = []
        # Scan all files under the path, And compare with the files recorded in self.previous_work_dir_files，
        # If there is a new file，Or the modification time of the file has changed，Then update in self.previous_work_dir_files
        # Record the paths of new files and files that have changed in change_list
        for root, dirs, files in os.walk(path_to_overwatch):
            for file in files:
                file_path = os.path.join(root, file)
                if file_path not in self.previous_work_dir_files.keys():
                    last_modified_time = os.stat(file_path).st_mtime
                    self.previous_work_dir_files.update({file_path: last_modified_time})
                    change_list.append(file_path)
                else:
                    last_modified_time = os.stat(file_path).st_mtime
                    if last_modified_time != self.previous_work_dir_files[file_path]:
                        self.previous_work_dir_files[file_path] = last_modified_time
                        change_list.append(file_path)
        if len(change_list) > 0:
            file_links = ""
            for f in change_list:
                res = promote_file_to_downloadzone(f)
                file_links += f'<br/><a href="file={res}" target="_blank">{res}</a>'
                yield from self.immediate_showoff_when_possible(f)

            self.chatbot.append(['Detect new documents.', f'The document list is as follows: {file_links}'])
            yield from update_ui(chatbot=self.chatbot, history=self.history)
        return change_list


    def main_process_ui_control(self, txt, create_or_resume) -> str:
        # ⭐ Main process
        if create_or_resume == 'create':
            self.cnt = 1
            self.parent_conn = self.launch_subprocess_with_pipe() # ⭐⭐⭐
        repeated, cmd_to_autogen = self.send_command(txt)
        if txt == 'exit': 
            self.chatbot.append([f"End", "The end signal is clear，Terminate AutoGen program。"])
            yield from update_ui(chatbot=self.chatbot, history=self.history)
            self.terminate()
            return "terminate"
        
        # patience = 10
        
        while True:
            time.sleep(0.5)
            if not self.alive:
                # the heartbeat watchdog might have it killed
                self.terminate()
                return "terminate"
            if self.parent_conn.poll(): 
                self.feed_heartbeat_watchdog()
                if "[GPT-Academic] Waiting" in self.chatbot[-1][-1]:
                    self.chatbot.pop(-1)  # remove the last line
                if "Waiting for your further instructions" in self.chatbot[-1][-1]:
                    self.chatbot.pop(-1)  # remove the last line
                if '[GPT-Academic] Waiting' in self.chatbot[-1][-1]:
                    self.chatbot.pop(-1)    # remove the last line
                msg = self.parent_conn.recv() # PipeCom
                if msg.cmd == "done":
                    self.chatbot.append([f"End", msg.content])
                    self.cnt += 1
                    yield from update_ui(chatbot=self.chatbot, history=self.history)
                    self.terminate()
                    break
                if msg.cmd == "show":
                    yield from self.overwatch_workdir_file_change()
                    notice = ""
                    if repeated: notice = "（Automatically ignore duplicate inputs）"
                    self.chatbot.append([f"Running phase-{self.cnt}（Last user feedback input is: 「{cmd_to_autogen}」{notice}", msg.content])
                    self.cnt += 1
                    yield from update_ui(chatbot=self.chatbot, history=self.history)
                if msg.cmd == "interact":
                    yield from self.overwatch_workdir_file_change()
                    self.chatbot.append([f"The program reaches the user feedback node.", msg.content + 
                                         "\n\nWaiting for your further instructions." + 
                                         "\n\n(1) In general, you don`t need to say anything, Clear input area, Then click `Submit` to continue. " +
                                         "\n\n(2) If you need to add something, Enter the content to be feedbacked, Click `Submit` to continue. " +
                                         "\n\n(3) If you want to terminate the program, Enter exit, Click `Submit` to terminate AutoGen and unlock. "
                    ])
                    yield from update_ui(chatbot=self.chatbot, history=self.history)
                    # do not terminate here, leave the subprocess_worker instance alive
                    return "wait_feedback"
            else:
                self.feed_heartbeat_watchdog()
                if '[GPT-Academic] Waiting' not in self.chatbot[-1][-1]:
                    # begin_waiting_time = time.time()
                    self.chatbot.append(["[GPT-Academic] Waiting for AutoGen execution result ...", "[GPT-Academic] Waiting"])
                self.chatbot[-1] = [self.chatbot[-1][0], self.chatbot[-1][1].replace("[GPT-Academic] Waiting", "[GPT-Academic] Waiting.")]
                yield from update_ui(chatbot=self.chatbot, history=self.history)
                # if time.time() - begin_waiting_time > patience:
                #     self.chatbot.append([f"End", "Waiting超When, Terminate AutoGen program。"])
                #     yield from update_ui(chatbot=self.chatbot, history=self.history)
                #     self.terminate()
                #     return "terminate"

        self.terminate()
        return "terminate"

    def subprocess_worker_wait_user_feedback(self, wait_msg="wait user feedback"):
        # ⭐⭐ run in subprocess
        patience = 5 * 60
        begin_waiting_time = time.time()
        self.child_conn.send(PipeCom("interact", wait_msg))
        while True:
            time.sleep(0.5)
            if self.child_conn.poll():
                wait_success = True
                break
            if time.time() - begin_waiting_time > patience:
                self.child_conn.send(PipeCom("done", ""))
                wait_success = False
                break
        return wait_success



from void_terminal.toolbox import CatchException, update_ui, gen_time_str
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive
from void_terminal.crazy_functions.crazy_utils import input_clipping
import copy, json

@CatchException
def 命令line助手(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    """
    txt             Text entered by the user in the input field, For example, a paragraph that needs to be translated, For example, a file path that contains files to be processed
    llm_kwargs      GPT model parameters, Such as temperature and top_p, Generally pass it on as is
    plugin_kwargs   Plugin model parameters, No use for the time being
    chatbot         Chat display box handle, Displayed to the user
    history         Chat history, Context summary
    system_prompt   Silent reminder to GPT
    web_port        Current software running port number
    """
    # Clear history, To avoid input overflow
    history = []

    # Input
    i_say = "Please write a bash command to implement the following function：" + txt
    # Start
    gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
        inputs=i_say, inputs_show_user=txt, 
        llm_kwargs=llm_kwargs, chatbot=chatbot, history=[], 
        sys_prompt="You are a Linux master-level user。Attention，When I ask you to write a bash command，Try to solve my request using only one command。"
    )
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page # UI update






from void_terminal.toolbox import update_ui
from void_terminal.toolbox import CatchException, report_exception
from void_terminal.toolbox import write_history_to_file, promote_file_to_downloadzone
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive
fast_debug = False

def GenerateFunctionComments(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt):
    import time, os
    print('begin analysis on:', file_manifest)
    for index, fp in enumerate(file_manifest):
        with open(fp, 'r', encoding='utf-8', errors='replace') as f:
            file_content = f.read()

        i_say = f'Please provide an overview of the program file below，And generate comments for all functions in the file，Output the results using markdown tables，The file name is{os.path.relpath(fp, project_folder)}，The file content is ```{file_content}```'
        i_say_show_user = f'[{index}/{len(file_manifest)}] Please provide an overview of the program file below，And generate comments for all functions in the file: {os.path.abspath(fp)}'
        chatbot.append((i_say_show_user, "[Local Message] waiting gpt response."))
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

        if not fast_debug: 
            msg = 'Normal'
            # ** gpt request **
            gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
                i_say, i_say_show_user, llm_kwargs, chatbot, history=[], sys_prompt=system_prompt)   # With timeout countdown

            chatbot[-1] = (i_say_show_user, gpt_say)
            history.append(i_say_show_user); history.append(gpt_say)
            yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page
            if not fast_debug: time.sleep(2)

    if not fast_debug: 
        res = write_history_to_file(history)
        promote_file_to_downloadzone(res, chatbot=chatbot)
        chatbot.append(("Are you done?？", res))
        yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page



@CatchException
def BatchGenerateFunctionComments(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    history = []    # Clear history，To avoid input overflow
    import glob, os
    if os.path.exists(txt):
        project_folder = txt
    else:
        if txt == "": txt = 'Empty input field'
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find local project or do not have access: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    file_manifest = [f for f in glob.glob(f'{project_folder}/**/*.py', recursive=True)] + \
                    [f for f in glob.glob(f'{project_folder}/**/*.cpp', recursive=True)]

    if len(file_manifest) == 0:
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find any .tex files: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    yield from GenerateFunctionComments(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt)



"""
https://github.com/langchain-ai/langchain/blob/master/docs/extras/modules/model_io/output_parsers/pydantic.ipynb

Example 1.

# Define your desired data structure.
class Joke(BaseModel):
    setup: str = Field(description="question to set up a joke")
    punchline: str = Field(description="answer to resolve the joke")

    # You can add custom validation logic easily with Pydantic.
    @validator("setup")
    def question_ends_with_question_mark(cls, field):
        if field[-1] != "?":
            raise ValueError("Badly formed question!")
        return field


Example 2.

# Here's another example, but with a compound typed field.
class Actor(BaseModel):
    name: str = Field(description="name of an actor")
    film_names: List[str] = Field(description="list of names of films they starred in")
"""

import json, re, logging


PYDANTIC_FORMAT_INSTRUCTIONS = """The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{schema}
```"""


PYDANTIC_FORMAT_INSTRUCTIONS_SIMPLE = """The output should be formatted as a JSON instance that conforms to the JSON schema below.
```
{schema}
```"""

class JsonStringError(Exception): ...

class GptJsonIO():

    def __init__(self, schema, example_instruction=True):
        self.pydantic_object = schema
        self.example_instruction = example_instruction
        self.format_instructions = self.generate_format_instructions()

    def generate_format_instructions(self):
        schema = self.pydantic_object.schema()

        # Remove extraneous fields.
        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        # Ensure json in context is well-formed with double quotes.
        if self.example_instruction:
            schema_str = json.dumps(reduced_schema)
            return PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)
        else:
            return PYDANTIC_FORMAT_INSTRUCTIONS_SIMPLE.format(schema=schema_str)

    def generate_output(self, text):
        # Greedy search for 1st json candidate.
        match = re.search(
            r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL
        )
        json_str = ""
        if match: json_str = match.group()
        json_object = json.loads(json_str, strict=False)
        final_object = self.pydantic_object.parse_obj(json_object)
        return final_object

    def generate_repair_prompt(self, broken_json, error):
        prompt = "Fix a broken json string.\n\n" + \
                 "(1) The broken json string need to fix is: \n\n" + \
                 "```" + "\n" + \
                 broken_json + "\n" + \
                 "```" + "\n\n" + \
                 "(2) The error message is: \n\n" + \
                 error + "\n\n" + \
                "Now, fix this json string. \n\n"
        return prompt
    
    def generate_output_auto_repair(self, response, gpt_gen_fn):
        """
        response: string containing canidate json
        gpt_gen_fn: gpt_gen_fn(inputs, sys_prompt)
        """
        try:
            result = self.generate_output(response)
        except Exception as e:
            try:
                logging.info(f'Repairing json：{response}')
                repair_prompt = self.generate_repair_prompt(broken_json = response, error=repr(e))
                result = self.generate_output(gpt_gen_fn(repair_prompt, self.format_instructions))
                logging.info('Repaire json success.')
            except Exception as e:
                # No way，Give up treatment
                logging.info('Repaire json fail.')
                raise JsonStringError('Cannot repair json.', str(e))
        return result




from void_terminal.toolbox import update_ui, get_log_folder
from void_terminal.toolbox import write_history_to_file, promote_file_to_downloadzone
from void_terminal.toolbox import CatchException, report_exception, get_conf
import re, requests, unicodedata, os
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive
def download_arxiv_(url_pdf):
    if 'arxiv.org' not in url_pdf:
        if ('.' in url_pdf) and ('/' not in url_pdf):
            new_url = 'https://arxiv.org/abs/'+url_pdf
            print('Download number：', url_pdf, 'Auto-locating：', new_url)
            # download_arxiv_(new_url)
            return download_arxiv_(new_url)
        else:
            print('Unrecognized URL!')
            return None
    if 'abs' in url_pdf:
        url_pdf = url_pdf.replace('abs', 'pdf')
        url_pdf = url_pdf + '.pdf'

    url_abs = url_pdf.replace('.pdf', '').replace('pdf', 'abs')
    title, other_info = get_name(_url_=url_abs)

    paper_id = title.split()[0]  # '[1712.00559]'
    if '2' in other_info['year']:
        title = other_info['year'] + ' ' + title

    known_conf = ['NeurIPS', 'NIPS', 'Nature', 'Science', 'ICLR', 'AAAI']
    for k in known_conf:
        if k in other_info['comment']:
            title = k + ' ' + title

    download_dir = get_log_folder(plugin_name='arxiv')
    os.makedirs(download_dir, exist_ok=True)

    title_str = title.replace('?', '？')\
        .replace(':', '：')\
        .replace('\"', '“')\
        .replace('\n', '')\
        .replace('  ', ' ')\
        .replace('  ', ' ')

    requests_pdf_url = url_pdf
    file_path = download_dir+title_str

    print('Downloading')
    proxies = get_conf('proxies')
    r = requests.get(requests_pdf_url, proxies=proxies)
    with open(file_path, 'wb+') as f:
        f.write(r.content)
    print('Download complete')

    # print('Output下载命令：','aria2c -o \"%s\" %s'%(title_str,url_pdf))
    # subprocess.call('aria2c --all-proxy=\"172.18.116.150:11084\" -o \"%s\" %s'%(download_dir+title_str,url_pdf), shell=True)

    x = "%s  %s %s.bib" % (paper_id, other_info['year'], other_info['authors'])
    x = x.replace('?', '？')\
        .replace(':', '：')\
        .replace('\"', '“')\
        .replace('\n', '')\
        .replace('  ', ' ')\
        .replace('  ', ' ')
    return file_path, other_info


def get_name(_url_):
    import os
    from bs4 import BeautifulSoup
    print('Getting article name!')
    print(_url_)

    # arxiv_recall = {}
    # if os.path.exists('./arxiv_recall.pkl'):
    #     with open('./arxiv_recall.pkl', 'rb') as f:
    #         arxiv_recall = pickle.load(f)

    # if _url_ in arxiv_recall:
    #     print('在缓存中')
    #     return arxiv_recall[_url_]

    proxies = get_conf('proxies')
    res = requests.get(_url_, proxies=proxies)

    bs = BeautifulSoup(res.text, 'html.parser')
    other_details = {}

    # get year
    try:
        year = bs.find_all(class_='dateline')[0].text
        year = re.search(r'(\d{4})', year, re.M | re.I).group(1)
        other_details['year'] = year
        abstract = bs.find_all(class_='abstract mathjax')[0].text
        other_details['abstract'] = abstract
    except:
        other_details['year'] = ''
        print('Failed to get year')

    # get author
    try:
        authors = bs.find_all(class_='authors')[0].text
        authors = authors.split('Authors:')[1]
        other_details['authors'] = authors
    except:
        other_details['authors'] = ''
        print('Failed to get authors')

    # get comment
    try:
        comment = bs.find_all(class_='metatable')[0].text
        real_comment = None
        for item in comment.replace('\n', ' ').split('   '):
            if 'Comments' in item:
                real_comment = item
        if real_comment is not None:
            other_details['comment'] = real_comment
        else:
            other_details['comment'] = ''
    except:
        other_details['comment'] = ''
        print('Failed to get year')

    title_str = BeautifulSoup(
        res.text, 'html.parser').find('title').contents[0]
    print('Successfully retrieved：', title_str)
    # arxiv_recall[_url_] = (title_str+'.pdf', other_details)
    # with open('./arxiv_recall.pkl', 'wb') as f:
    #     pickle.dump(arxiv_recall, f)

    return title_str+'.pdf', other_details



@CatchException
def DownloadArxivPaperAndTranslateAbstract(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):

    CRAZY_FUNCTION_INFO = "DownloadArxivPaperAndTranslateAbstract，Function plugin author[binary-husky]。Extracting abstract and downloading PDF document..."
    import glob
    import os

    # Basic information：Function, contributor
    chatbot.append(["Function plugin feature？", CRAZY_FUNCTION_INFO])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # Attempt to import dependencies，If dependencies are missing，Give installation suggestions
    try:
        import bs4
    except:
        report_exception(chatbot, history, 
            a = f"Parsing project: {txt}", 
            b = f"Failed to import software dependencies。Using this module requires additional dependencies，Installation method```pip install --upgrade beautifulsoup4```。")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return

    # Clear history，To avoid input overflow
    history = []

    # Extract abstract，Download PDF document
    try:
        pdf_path, info = download_arxiv_(txt)
    except:
        report_exception(chatbot, history, 
            a = f"Parsing project: {txt}", 
            b = f"PDF file download unsuccessful")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    
    # Translate abstract, etc.
    i_say =            f"Please read the following academic paper related materials，Extract abstract，Translate to Chinese。Materials are as follows：{str(info)}"
    i_say_show_user =  f'Please read the following academic paper related materials，Extract abstract，Translate to Chinese。Paper：{pdf_path}'
    chatbot.append((i_say_show_user, "[Local Message] waiting gpt response."))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
    msg = 'Normal'
    # ** gpt request **
    # Single line，Get article meta information
    gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
        inputs=i_say,
        inputs_show_user=i_say_show_user,
        llm_kwargs=llm_kwargs,
        chatbot=chatbot, history=[],
        sys_prompt="Your job is to collect information from materials and translate to Chinese。",
    )

    chatbot[-1] = (i_say_show_user, gpt_say)
    history.append(i_say_show_user); history.append(gpt_say)
    yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page
    res = write_history_to_file(history)
    promote_file_to_downloadzone(res, chatbot=chatbot)
    promote_file_to_downloadzone(pdf_path, chatbot=chatbot)

    chatbot.append(("Are you done?？", res + "\n\nPDF file has also been downloaded"))
    yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page




sample = """
[1]: https://baike.baidu.com/item/%E8%B4%A8%E8%83%BD%E6%96%B9%E7%A8%8B/1884527 "质能方程（Mass-energy equivalence equation）_百度百科"
[2]: https://www.zhihu.com/question/348249281 " e.g., 何理解质能方程 E＝mc²？ - Zhihu"
[3]: https://zhuanlan.zhihu.com/p/32597385 "质能方程的推导与理解 - Zhihu - Zhihu专栏"

Hello，This is Bing。The mass-energy equivalence equation describes the equivalent relationship between mass and energy[^1^][1]。In tex format，The mass-energy equivalence equation can be written as $$E=mc^2$$，Where $E$ is energy，$m$ is mass，$c$ is the speed of light[^2^][2] [^3^][3]。
"""
import re

def preprocess_newbing_out(s):
    pattern = r'\^(\d+)\^' # Match ^number^
    pattern2 = r'\[(\d+)\]' # Match ^number^
    sub = lambda m: '\['+m.group(1)+'\]' # Replace the matched number as the replacement value
    result = re.sub(pattern, sub, s) # Replacement operation
    if '[1]' in result:
        result += '<br/><hr style="border-top: dotted 1px #44ac5c;"><br/><small>' + "<br/>".join([re.sub(pattern2, sub, r) for r in result.split('\n') if r.startswith('[')]) + '</small>'
    return result


def close_up_code_segment_during_stream(gpt_reply):
    """
    In the middle of outputting code with GPT（Output the front part```，But haven`t output the back part yet```），Complete the back part```

    Args:
        gpt_reply (str): Reply string returned by GPT model。

    Returns:
        str: Return a new string，Append the back part of output code snippet```to it。

    """
    if '```' not in gpt_reply:
        return gpt_reply
    if gpt_reply.endswith('```'):
        return gpt_reply

    # Exclude the above two cases，We
    segments = gpt_reply.split('```')
    n_mark = len(segments) - 1
    if n_mark % 2 == 1:
        # print('Output代码Segment中！')
        return gpt_reply+'\n```'
    else:
        return gpt_reply
    
import markdown
from latex2mathml.converter import convert as tex2mathml
from functools import wraps, lru_cache
def markdown_convertion(txt):
    """
    Convert Markdown format text to HTML format。If it contains mathematical formulas，Convert the formula to HTML format first。
    """
    pre = '<div class="markdown-body">'
    suf = '</div>'
    if txt.startswith(pre) and txt.endswith(suf):
        # print('Warning，Input a string that has already been converted，二次转化可能出Question')
        return txt # Has already been converted，No need to convert again
    
    markdown_extension_configs = {
        'mdx_math': {
            'enable_dollar_delimiter': True,
            'use_gitlab_delimiters': False,
        },
    }
    find_equation_pattern = r'<script type="math/tex(?:.*?)>(.*?)</script>'

    def tex2mathml_catch_exception(content, *args, **kwargs):
        try:
            content = tex2mathml(content, *args, **kwargs)
        except:
            content = content
        return content

    def replace_math_no_render(match):
        content = match.group(1)
        if 'mode=display' in match.group(0):
            content = content.replace('\n', '</br>')
            return f"<font color=\"#00FF00\">$$</font><font color=\"#FF00FF\">{content}</font><font color=\"#00FF00\">$$</font>"
        else:
            return f"<font color=\"#00FF00\">$</font><font color=\"#FF00FF\">{content}</font><font color=\"#00FF00\">$</font>"

    def replace_math_render(match):
        content = match.group(1)
        if 'mode=display' in match.group(0):
            if '\\begin{aligned}' in content:
                content = content.replace('\\begin{aligned}', '\\begin{array}')
                content = content.replace('\\end{aligned}', '\\end{array}')
                content = content.replace('&', ' ')
            content = tex2mathml_catch_exception(content, display="block")
            return content
        else:
            return tex2mathml_catch_exception(content)

    def markdown_bug_hunt(content):
        """
        Fix a bug in mdx_math（Redundant when wrapping begin command with single $<script>）
        """
        content = content.replace('<script type="math/tex">\n<script type="math/tex; mode=display">', '<script type="math/tex; mode=display">')
        content = content.replace('</script>\n</script>', '</script>')
        return content


    if ('$' in txt) and ('```' not in txt):  # Formula symbol with $ sign，And there is no code section```Identifier of
        # convert everything to html format
        split = markdown.markdown(text='---')
        convert_stage_1 = markdown.markdown(text=txt, extensions=['mdx_math', 'fenced_code', 'tables', 'sane_lists'], extension_configs=markdown_extension_configs)
        convert_stage_1 = markdown_bug_hunt(convert_stage_1)
        # re.DOTALL: Make the '.' special character match any character at all, including a newline; without this flag, '.' will match anything except a newline. Corresponds to the inline flag (?s).
        # 1. convert to easy-to-copy tex (do not render math)
        convert_stage_2_1, n = re.subn(find_equation_pattern, replace_math_no_render, convert_stage_1, flags=re.DOTALL)
        # 2. convert to rendered equation
        convert_stage_2_2, n = re.subn(find_equation_pattern, replace_math_render, convert_stage_1, flags=re.DOTALL)
        # cat them together
        return pre + convert_stage_2_1 + f'{split}' + convert_stage_2_2 + suf
    else:
        return pre + markdown.markdown(txt, extensions=['fenced_code', 'codehilite', 'tables', 'sane_lists']) + suf


sample = preprocess_newbing_out(sample)
sample = close_up_code_segment_during_stream(sample)
sample = markdown_convertion(sample)
with open('tmp.html', 'w', encoding='utf8') as f:
    f.write("""

<head>
    <title>My Website</title>
    <link rel="stylesheet" type="text/css" href="style.css">
</head>

    """)
    f.write(sample)



from void_terminal.toolbox import update_ui
from void_terminal.toolbox import CatchException, report_exception
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive
from void_terminal.toolbox import write_history_to_file, promote_file_to_downloadzone

fast_debug = False

def readPdf(pdfPath):
    """
    Read the pdf file，Return the text content
    """
    import pdfminer
    from pdfminer.pdfparser import PDFParser
    from pdfminer.pdfdocument import PDFDocument
    from pdfminer.pdfpage import PDFPage, PDFTextExtractionNotAllowed
    from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
    from pdfminer.pdfdevice import PDFDevice
    from pdfminer.layout import LAParams
    from pdfminer.converter import PDFPageAggregator

    fp = open(pdfPath, 'rb')

    # Create a PDF parser object associated with the file object
    parser = PDFParser(fp)

    # Create a PDF document object that stores the document structure.
    # Password for initialization as 2nd parameter
    document = PDFDocument(parser)
    # Check if the document allows text extraction. If not, abort.
    if not document.is_extractable:
        raise PDFTextExtractionNotAllowed

    # Create a PDF resource manager object that stores shared resources.
    rsrcmgr = PDFResourceManager()

    # Create a PDF device object.
    # device = PDFDevice(rsrcmgr)

    # BEGIN LAYOUT ANALYSIS.
    # Set parameters for analysis.
    laparams = LAParams(
        char_margin=10.0,
        line_margin=0.2,
        boxes_flow=0.2,
        all_texts=False,
    )
    # Create a PDF page aggregator object.
    device = PDFPageAggregator(rsrcmgr, laparams=laparams)
    # Create a PDF interpreter object.
    interpreter = PDFPageInterpreter(rsrcmgr, device)

    # loop over all pages in the document
    outTextList = []
    for page in PDFPage.create_pages(document):
        # read the page into a layout object
        interpreter.process_page(page)
        layout = device.get_result()
        for obj in layout._objs:
            if isinstance(obj, pdfminer.layout.LTTextBoxHorizontal):
                # print(obj.get_text())
                outTextList.append(obj.get_text())

    return outTextList


def ParsePaper(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt):
    import time, glob, os
    from bs4 import BeautifulSoup
    print('begin analysis on:', file_manifest)
    for index, fp in enumerate(file_manifest):
        if ".tex" in fp:
            with open(fp, 'r', encoding='utf-8', errors='replace') as f:
                file_content = f.read()
        if ".pdf" in fp.lower():
            file_content = readPdf(fp)
            file_content = BeautifulSoup(''.join(file_content), features="lxml").body.text.encode('gbk', 'ignore').decode('gbk')

        prefix = "Next, please analyze the following paper files one by one，Summarize its content" if index==0 else ""
        i_say = prefix + f'Please summarize the following article in Chinese，The file name is{os.path.relpath(fp, project_folder)}，The content of the article is ```{file_content}```'
        i_say_show_user = prefix + f'[{index}/{len(file_manifest)}] Please summarize the following article: {os.path.abspath(fp)}'
        chatbot.append((i_say_show_user, "[Local Message] waiting gpt response."))
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

        if not fast_debug:
            msg = 'Normal'
            # ** gpt request **
            gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
                inputs=i_say, 
                inputs_show_user=i_say_show_user, 
                llm_kwargs=llm_kwargs,
                chatbot=chatbot, 
                history=[],
                sys_prompt="Summarize the article。"
            )  # With timeout countdown
            chatbot[-1] = (i_say_show_user, gpt_say)
            history.append(i_say_show_user); history.append(gpt_say)
            yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page
            if not fast_debug: time.sleep(2)

    all_file = ', '.join([os.path.relpath(fp, project_folder) for index, fp in enumerate(file_manifest)])
    i_say = f'According to your own analysis above，Summarize the entire text，Write a Chinese abstract in academic language，Then write an English abstract（Including{all_file}）。'
    chatbot.append((i_say, "[Local Message] waiting gpt response."))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    if not fast_debug:
        msg = 'Normal'
        # ** gpt request **
        gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
            inputs=i_say, 
            inputs_show_user=i_say, 
            llm_kwargs=llm_kwargs,
            chatbot=chatbot, 
            history=history,
            sys_prompt="Summarize the article。"
        )  # With timeout countdown
        chatbot[-1] = (i_say, gpt_say)
        history.append(i_say); history.append(gpt_say)
        yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page
        res = write_history_to_file(history)
        promote_file_to_downloadzone(res, chatbot=chatbot)
        chatbot.append(("Are you done?？", res))
        yield from update_ui(chatbot=chatbot, history=history, msg=msg) # Refresh the page



@CatchException
def BatchSummarizePDFDocumentspdfminer(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    history = []    # Clear history，To avoid input overflow
    import glob, os

    # Basic information：Function, contributor
    chatbot.append([
        "Function plugin feature？",
        "BatchSummarizePDFDocuments，This version uses the pdfminer plugin，With token reduction function。Function plugin contributor: Euclid-Jie。"])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # Attempt to import dependencies，If dependencies are missing，Give installation suggestions
    try:
        import pdfminer, bs4
    except:
        report_exception(chatbot, history, 
            a = f"Parsing project: {txt}", 
            b = f"Failed to import software dependencies。Using this module requires additional dependencies，Installation method```pip install --upgrade pdfminer beautifulsoup4```。")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    if os.path.exists(txt):
        project_folder = txt
    else:
        if txt == "": txt = 'Empty input field'
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find local project or do not have access: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    file_manifest = [f for f in glob.glob(f'{project_folder}/**/*.tex', recursive=True)] + \
                    [f for f in glob.glob(f'{project_folder}/**/*.pdf', recursive=True)] # + \
                    # [f for f in glob.glob(f'{project_folder}/**/*.cpp', recursive=True)] + \
                    # [f for f in glob.glob(f'{project_folder}/**/*.c', recursive=True)]
    if len(file_manifest) == 0:
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find any .tex or .pdf files: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    yield from ParsePaper(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt)





import time
from void_terminal.toolbox import update_ui, get_conf, update_ui_lastest_msg

model_name = 'Zhipu AI large model'

def validate_key():
    ZHIPUAI_API_KEY = get_conf("ZHIPUAI_API_KEY")
    if ZHIPUAI_API_KEY == '': return False
    return True

def predict_no_ui_long_connection(inputs, llm_kwargs, history=[], sys_prompt="", observe_window=[], console_slience=False):
    """
        ⭐Multi-threaded method
        For function details, please see request_llms/bridge_all.py
    """
    watch_dog_patience = 5
    response = ""

    if validate_key() is False:
        raise RuntimeError('Please configure ZHIPUAI_API_KEY')

    from void_terminal.request_llms.com_zhipuapi import ZhipuRequestInstance
    sri = ZhipuRequestInstance()
    for response in sri.generate(inputs, llm_kwargs, history, sys_prompt):
        if len(observe_window) >= 1:
            observe_window[0] = response
        if len(observe_window) >= 2:
            if (time.time()-observe_window[1]) > watch_dog_patience: raise RuntimeError("Program terminated。")
    return response

def predict(inputs, llm_kwargs, plugin_kwargs, chatbot, history=[], system_prompt='', stream = True, additional_fn=None):
    """
        ⭐Single-threaded method
        For function details, please see request_llms/bridge_all.py
    """
    chatbot.append((inputs, ""))
    yield from update_ui(chatbot=chatbot, history=history)

    if validate_key() is False:
        yield from update_ui_lastest_msg(lastmsg="[Local Message] Please configure ZHIPUAI_API_KEY", chatbot=chatbot, history=history, delay=0)
        return

    if additional_fn is not None:
        from void_terminal.core_functional import handle_core_functionality
        inputs, history = handle_core_functionality(additional_fn, inputs, history, chatbot)

    # Start receiving replies    
    from void_terminal.request_llms.com_zhipuapi import ZhipuRequestInstance
    sri = ZhipuRequestInstance()
    for response in sri.generate(inputs, llm_kwargs, history, system_prompt):
        chatbot[-1] = (inputs, response)
        yield from update_ui(chatbot=chatbot, history=history)

    # Summary output
    if response == f"[Local Message] Waiting{model_name}Responding ...":
        response = f"[Local Message] {model_name}Response exception ..."
    history.extend([inputs, response])
    yield from update_ui(chatbot=chatbot, history=history)


from void_terminal.toolbox import update_ui, trimmed_format_exc, promote_file_to_downloadzone, get_log_folder
from void_terminal.toolbox import CatchException, report_exception, write_history_to_file, zip_folder


class PaperFileGroup():
    def __init__(self):
        self.file_paths = []
        self.file_contents = []
        self.sp_file_contents = []
        self.sp_file_index = []
        self.sp_file_tag = []

        # count_token
        from void_terminal.request_llms.bridge_all import model_info
        enc = model_info["gpt-3.5-turbo"]['tokenizer']
        def get_token_num(txt): return len(enc.encode(txt, disallowed_special=()))
        self.get_token_num = get_token_num

    def run_file_split(self, max_token_limit=1900):
        """
        Separate long text
        """
        for index, file_content in enumerate(self.file_contents):
            if self.get_token_num(file_content) < max_token_limit:
                self.sp_file_contents.append(file_content)
                self.sp_file_index.append(index)
                self.sp_file_tag.append(self.file_paths[index])
            else:
                from void_terminal.crazy_functions.crazy_utils import breakdown_txt_to_satisfy_token_limit_for_pdf
                segments = breakdown_txt_to_satisfy_token_limit_for_pdf(file_content, self.get_token_num, max_token_limit)
                for j, segment in enumerate(segments):
                    self.sp_file_contents.append(segment)
                    self.sp_file_index.append(index)
                    self.sp_file_tag.append(self.file_paths[index] + f".part-{j}.tex")

        print('Segmentation: done')
    def merge_result(self):
        self.file_result = ["" for _ in range(len(self.file_paths))]
        for r, k in zip(self.sp_file_result, self.sp_file_index):
            self.file_result[k] += r

    def write_result(self):
        manifest = []
        for path, res in zip(self.file_paths, self.file_result):
            with open(path + '.polish.tex', 'w', encoding='utf8') as f:
                manifest.append(path + '.polish.tex')
                f.write(res)
        return manifest
    
    def zip_result(self):
        import os, time
        folder = os.path.dirname(self.file_paths[0])
        t = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime())
        zip_folder(folder, get_log_folder(), f'{t}-polished.zip')


def ProofreadMultipleFiles(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, language='en', mode='polish'):
    import time, os, re
    from void_terminal.crazy_functions.crazy_utils import request_gpt_model_multi_threads_with_very_awesome_ui_and_high_efficiency


    #  <-------- Read Latex file，Remove all comments from it ----------> 
    pfg = PaperFileGroup()

    for index, fp in enumerate(file_manifest):
        with open(fp, 'r', encoding='utf-8', errors='replace') as f:
            file_content = f.read()
            # Define the regular expression of comments
            comment_pattern = r'(?<!\\)%.*'
            # Use regular expressions to find comments，And replace them with an empty string
            clean_tex_content = re.sub(comment_pattern, '', file_content)
            # Record the text after removing comments
            pfg.file_paths.append(fp)
            pfg.file_contents.append(clean_tex_content)

    #  <-------- Split long latex files ----------> 
    pfg.run_file_split(max_token_limit=1024)
    n_split = len(pfg.sp_file_contents)


    #  <-------- Multithreading polishing begins ----------> 
    if language == 'en':
        if mode == 'polish':
            inputs_array = ["Below is a section from an academic paper, polish this section to meet the academic standard, " + 
                            "improve the grammar, clarity and overall readability, do not modify any latex command such as \section, \cite and equations:" + 
                            f"\n\n{frag}" for frag in pfg.sp_file_contents]
        else:
            inputs_array = [r"Below is a section from an academic paper, proofread this section." + 
                            r"Do not modify any latex command such as \section, \cite, \begin, \item and equations. " + 
                            r"Answer me only with the revised text:" + 
                        f"\n\n{frag}" for frag in pfg.sp_file_contents]
        inputs_show_user_array = [f"Polish {f}" for f in pfg.sp_file_tag]
        sys_prompt_array = ["You are a professional academic paper writer." for _ in range(n_split)]
    elif language == 'zh':
        if mode == 'polish':
            inputs_array = [f"The following is a paragraph from an academic paper，Please polish this section to meet academic standards，Improve grammar, clarity, and overall readability，Do not modify any LaTeX commands，such as \section，\cite and equations：" + 
                            f"\n\n{frag}" for frag in pfg.sp_file_contents]
        else:
            inputs_array = [f"The following is a paragraph from an academic paper，Please correct the grammar of this part。Do not modify any LaTeX commands，such as \section，\cite and equations：" + 
                            f"\n\n{frag}" for frag in pfg.sp_file_contents] 
        inputs_show_user_array = [f"Polishing {f}" for f in pfg.sp_file_tag]
        sys_prompt_array=["You are a professional Chinese academic paper writer。" for _ in range(n_split)]


    gpt_response_collection = yield from request_gpt_model_multi_threads_with_very_awesome_ui_and_high_efficiency(
        inputs_array=inputs_array,
        inputs_show_user_array=inputs_show_user_array,
        llm_kwargs=llm_kwargs,
        chatbot=chatbot,
        history_array=[[""] for _ in range(n_split)],
        sys_prompt_array=sys_prompt_array,
        # max_workers=5,  # Parallel task number limit，Up to 5 can be executed at the same time，Others are queued and waiting
        scroller_max_len = 80
    )

    #  <-------- Reassemble text fragments into a complete tex file，Organize the results into a compressed package ----------> 
    try:
        pfg.sp_file_result = []
        for i_say, gpt_say in zip(gpt_response_collection[0::2], gpt_response_collection[1::2]):
            pfg.sp_file_result.append(gpt_say)
        pfg.merge_result()
        pfg.write_result()
        pfg.zip_result()
    except:
        print(trimmed_format_exc())

    #  <-------- Organize the results，Exit ----------> 
    create_report_file_name = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime()) + f"-chatgpt.polish.md"
    res = write_history_to_file(gpt_response_collection, file_basename=create_report_file_name)
    promote_file_to_downloadzone(res, chatbot=chatbot)

    history = gpt_response_collection
    chatbot.append((f"{fp}Are you done?？", res))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page


@CatchException
def EnglishProofreadingForLatex(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    # Basic information：Function, contributor
    chatbot.append([
        "Function plugin feature？",
        "Polish the entire Latex project。Function plugin contributor: Binary-Husky。（Attention，This plugin does not call Latex，If there is a Latex environment，Please use the `LatexEnglishCorrection+Highlight` plugin）"])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # Attempt to import dependencies，If dependencies are missing，Give installation suggestions
    try:
        import tiktoken
    except:
        report_exception(chatbot, history,
                         a=f"Parsing project: {txt}",
                         b=f"Failed to import software dependencies。Using this module requires additional dependencies，Installation method```pip install --upgrade tiktoken```。")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    history = []    # Clear history，To avoid input overflow
    import glob, os
    if os.path.exists(txt):
        project_folder = txt
    else:
        if txt == "": txt = 'Empty input field'
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find local project or do not have access: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    file_manifest = [f for f in glob.glob(f'{project_folder}/**/*.tex', recursive=True)]
    if len(file_manifest) == 0:
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find any .tex files: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    yield from ProofreadMultipleFiles(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, language='en')






@CatchException
def LatexChineseProofreading(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    # Basic information：Function, contributor
    chatbot.append([
        "Function plugin feature？",
        "Polish the entire Latex project。Function plugin contributor: Binary-Husky"])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # Attempt to import dependencies，If dependencies are missing，Give installation suggestions
    try:
        import tiktoken
    except:
        report_exception(chatbot, history,
                         a=f"Parsing project: {txt}",
                         b=f"Failed to import software dependencies。Using this module requires additional dependencies，Installation method```pip install --upgrade tiktoken```。")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    history = []    # Clear history，To avoid input overflow
    import glob, os
    if os.path.exists(txt):
        project_folder = txt
    else:
        if txt == "": txt = 'Empty input field'
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find local project or do not have access: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    file_manifest = [f for f in glob.glob(f'{project_folder}/**/*.tex', recursive=True)]
    if len(file_manifest) == 0:
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find any .tex files: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    yield from ProofreadMultipleFiles(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, language='zh')




@CatchException
def LatexEnglishCorrection(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    # Basic information：Function, contributor
    chatbot.append([
        "Function plugin feature？",
        "Correcting the entire Latex project。Function plugin contributor: Binary-Husky"])
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    # Attempt to import dependencies，If dependencies are missing，Give installation suggestions
    try:
        import tiktoken
    except:
        report_exception(chatbot, history,
                         a=f"Parsing project: {txt}",
                         b=f"Failed to import software dependencies。Using this module requires additional dependencies，Installation method```pip install --upgrade tiktoken```。")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    history = []    # Clear history，To avoid input overflow
    import glob, os
    if os.path.exists(txt):
        project_folder = txt
    else:
        if txt == "": txt = 'Empty input field'
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find local project or do not have access: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    file_manifest = [f for f in glob.glob(f'{project_folder}/**/*.tex', recursive=True)]
    if len(file_manifest) == 0:
        report_exception(chatbot, history, a = f"Parsing project: {txt}", b = f"Cannot find any .tex files: {txt}")
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return
    yield from ProofreadMultipleFiles(file_manifest, project_folder, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, language='en', mode='proofread')






"""
    All the following configurations can also be overridden using environment variables，See docker-compose.yml for the format of environment variable configuration。
    Read priority：Environment variable > config_private.py > config.py
    --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
    All the following configurations also support using environment variables to override, 
    and the environment variable configuration format can be seen in docker-compose.yml. 
    Configuration reading priority: environment variable > config_private.py > config.py
"""

# [step 1]>> API_KEY = "sk-123456789xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx123456789"。In very few cases，You also need to fill in the organization（In the format of org-123456789abcdefghijklmno），Please scroll down，Find API_ORG setting item
API_KEY = "Fill in the API key here"    # Multiple API-KEYs can be filled in at the same time，Separated by commas，For exampleAPI_KEY = "sk-openaikey1,sk-openaikey2,fkxxxx-api2dkey3,azure-apikey4"


# [step 2]>> Change to True to apply proxy，If deployed directly on overseas servers，Do not modify here; if using local or region-unrestricted large models，No modification is needed here
USE_PROXY = False
if USE_PROXY:
    """
    Format for filling in is [Protocol]://  [Address] :[Port]，Don`t forget to change USE_PROXY to True before filling in，If deployed directly on overseas servers，Do not modify here
            <Configuration tutorial & video tutorial> https://github.com/binary-husky/gpt_academic/issues/1>
    [Protocol] Common protocols are nothing but socks5h/http; For example, the default local protocol for v2**y and ss* is socks5h; While the default local protocol for cl**h is http
    [Address] Those who understand, understand，If you don`t understand, just fill in localhost or 127.0.0.1 and you won`t go wrong（localhost means that the proxy software is installed on the local machine）
    [Port] Look for it in the settings of the proxy software。Although the interface of different proxy software is different，But the port number should be in the most prominent position
    """
    # Address of the proxy network，Open your *learning* software to view the proxy protocol(socks5h / http)and address(localhost)and port(11284)
    proxies = {
        #          [Protocol]://  [Address]  :[Port]
        "http":  "socks5h://localhost:11284",  # 再For example  "http":  "http://127.0.0.1:7890",
        "https": "socks5h://localhost:11284",  # 再For example  "https": "http://127.0.0.1:7890",
    }
else:
    proxies = None

# ------------------------------------ The following configurations can optimize the experience, However, it does not need to be modified in most cases ------------------------------------

# Redirect URL，Realize the function of changing API_URL（High-risk setting! Do not modify under normal circumstances! Modify this setting，You will completely expose your API-KEY and conversation privacy to the middleman you set!）
# Format: API_URL_REDIRECT = {"https://api.openai.com/v1/chat/completions": "在这里填写重定To的api.openai.com的URL"} 
# For example: API_URL_REDIRECT = {"https://api.openai.com/v1/chat/completions": "https://reverse-proxy-url/v1/chat/completions"}
API_URL_REDIRECT = {}


# In the multi-threaded function plugin，How many threads are allowed to access OpenAI at the same time by default。The limit for free trial users is 3 times per minute，The limit for Pay-as-you-go users is 3500 times per minute
# In short：Free（5 dollars）User fills in 3，Users who have bound their credit card to OpenAI can fill in 16 or higher。Please check for higher limits：https://platform.openai.com/docs/guides/rate-limits/overview
DEFAULT_WORKER_NUM = 3


# Color theme, Optional ["Default", "Chuanhu-Small-and-Beautiful", "High-Contrast"]
# More themes, Please refer to the Gradio theme store: https://huggingface.co/spaces/gradio/theme-gallery optional ["Gstaff/Xkcd", "NoCrypt/Miku", ...]
THEME = "Default"
AVAIL_THEMES = ["Default", "Chuanhu-Small-and-Beautiful", "High-Contrast", "Gstaff/Xkcd", "NoCrypt/Miku"]


# Default system prompts（system prompt）
INIT_SYS_PROMPT = "Serve me as a writing and programming assistant."


# Height of the Dialogue Window （仅在LAYOUT="TOP-DOWN"When生效）
CHATBOT_HEIGHT = 1115


# Code Highlighting
CODE_HIGHLIGHT = True


# Window Layout
LAYOUT = "LEFT-RIGHT"   # "LEFT-RIGHT"（Horizontal Layout） # "TOP-DOWN"（Vertical Layout）


# Dark mode / Light mode
DARK_MODE = True        


# After sending the request to OpenAI，Timeout Threshold
TIMEOUT_SECONDS = 30


# Web Port, -1 represents random port
WEB_PORT = -1


# If OpenAI does not respond（Network Lag, Proxy Failure, KEY Invalid），Retry Limit
MAX_RETRY = 2


# Default options for plugin classification
DEFAULT_FN_GROUPS = ['Conversation', 'programming', 'Academic', 'Intelligent agent']


# Model selection is (Attention: LLM_MODEL is the default selected model, It *must* be included in the AVAIL_LLM_MODELS list )
LLM_MODEL = "gpt-3.5-turbo" # Optional ↓↓↓
AVAIL_LLM_MODELS = ["gpt-3.5-turbo-1106","gpt-4-1106-preview",
                    "gpt-3.5-turbo-16k", "gpt-3.5-turbo", "azure-gpt-3.5",
                    "api2d-gpt-3.5-turbo", 'api2d-gpt-3.5-turbo-16k',
                    "gpt-4", "gpt-4-32k", "azure-gpt-4", "api2d-gpt-4",
                    "chatglm3", "moss", "newbing", "claude-2"]
# P.S. Other available models include ["zhipuai", "qianfan", "llama2", "qwen", "gpt-3.5-turbo-0613", "gpt-3.5-turbo-16k-0613",  "gpt-3.5-random"
# "spark", "sparkv2", "sparkv3", "chatglm_onnx", "claude-1-100k", "claude-2", "internlm", "jittorllms_pangualpha", "jittorllms_llama"]


# Define which models the `Ask multiple GPT models` plugin should use on the interface，Please select from AVAIL_LLM_MODELS，And use it between different models`&`Interval，For example"gpt-3.5-turbo&chatglm3&azure-gpt-4"
MULTI_QUERY_LLM_MODELS = "gpt-3.5-turbo&chatglm3"


# Baidu Qianfan（LLM_MODEL="qianfan"）
BAIDU_CLOUD_API_KEY = ''
BAIDU_CLOUD_SECRET_KEY = ''
BAIDU_CLOUD_QIANFAN_MODEL = 'ERNIE-Bot'    # Optional "ERNIE-Bot-4"(Wenxin Large Model 4.0), "ERNIE-Bot"(Original text), "ERNIE-Bot-turbo", "BLOOMZ-7B", "Llama-2-70B-Chat", "Llama-2-13B-Chat", "Llama-2-7B-Chat"


# If using ChatGLM2 fine-tuning model，请把 LLM_MODEL="chatglmft"，And specify the model path here
CHATGLM_PTUNING_CHECKPOINT = "" # For example"/home/hmp/ChatGLM2-6B/ptuning/output/6b-pt-128-1e-2/checkpoint-100"


# Execution mode of local LLM models such as ChatGLM CPU/GPU
LOCAL_MODEL_DEVICE = "cpu" # Optional "cuda"
LOCAL_MODEL_QUANT = "FP16" # Default "FP16" "INT4" 启用量化INT4version本 "INT8" 启用量化INT8version本


# Set the number of parallel threads for Gradio（No modification is needed）
CONCURRENT_COUNT = 100


# Whether to automatically clear the input box upon submission
AUTO_CLEAR_TXT = False


# Add a Live2D decoration
ADD_WAIFU = False


# Set username and password（No modification is needed）（Related functions are unstable，Related to gradio version and network，Not recommended to add this for local use）
# [("username", "password"), ("username2", "password2"), ...]
AUTHENTICATION = []


# If you need to run under the second-level path（Under normal circumstances，Do not modify!!）（Need to be modified in conjunction with main.py to take effect!）
CUSTOM_PATH = "/"


# HTTPS Key and Certificate（No modification is needed）
SSL_KEYFILE = ""
SSL_CERTFILE = ""


# In very few cases，Openai`s official KEY needs to be accompanied by organizational code（Format like org-xxxxxxxxxxxxxxxxxxxxxxxx）Use
API_ORG = ""


# If you need to use Slack Claude，See request_llms/README.md for detailed usage instructions
SLACK_CLAUDE_BOT_ID = ''   
SLACK_CLAUDE_USER_TOKEN = ''


# If you need to use AZURE（Method 1：Deploy a single Azure model）For details, please refer to the additional document docs\use_azure.md
AZURE_ENDPOINT = "https://The API name you wrote yourself.openai.azure.com/"
AZURE_API_KEY = "Fill in the Azure OpenAI API key"    # It is recommended to fill in directly at API_KEY.，This option will be deprecated soon
AZURE_ENGINE = "Fill in the deployment name you wrote by yourself"            # Read docs\use_azure.md


# If you need to use AZURE（Method 2：Deploy multiple Azure models + dynamic switching）For details, please refer to the additional document docs\use_azure.md
AZURE_CFG_ARRAY = {}


# Using Newbing (Not recommended for use，Will be deleted in the future)
NEWBING_STYLE = "creative"  # ["creative", "balanced", "precise"]
NEWBING_COOKIES = """
put your new bing cookies here
"""


# Aliyun real-time speech recognition has high configuration difficulty and is only recommended for advanced users. Refer to https://github.com/binary-husky/gpt_academic/blob/master/docs/use_audio.md
ENABLE_AUDIO = False
ALIYUN_TOKEN=""     # For example, f37f30e0f9934c34a992f6f64f7eba4f
ALIYUN_APPKEY=""    # For example, RoPlZrM88DnAFkZK
ALIYUN_ACCESSKEY="" # （No need to fill in）
ALIYUN_SECRET=""    # （No need to fill in）


# Access to Xunfei Xinghuo large model https://console.xfyun.cn/services/iat
XFYUN_APPID = "00000000"
XFYUN_API_SECRET = "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
XFYUN_API_KEY = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"


# Access the intelligent spectrum model
ZHIPUAI_API_KEY = ""
ZHIPUAI_MODEL = "chatglm_turbo"


# Claude API KEY
ANTHROPIC_API_KEY = ""


# Customize API KEY format
CUSTOM_API_KEY_PATTERN = ""


# HUGGINGFACE`s TOKEN，Works when downloading LLAMA https://huggingface.co/docs/hub/security-tokens
HUGGINGFACE_ACCESS_TOKEN = "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"


# GROBID server address（Filling in multiple can balance the load），Used for high-quality reading of PDF documents
# Get method：Copy the following space https://huggingface.co/spaces/qingxu98/grobid，Set as public，然后GROBID_URL = "https://(Your hf username is qingxu98)-(Your filled space name such as grobid).hf.space"
GROBID_URLS = [
    "https://qingxu98-grobid.hf.space","https://qingxu98-grobid2.hf.space","https://qingxu98-grobid3.hf.space",
    "https://qingxu98-grobid4.hf.space","https://qingxu98-grobid5.hf.space", "https://qingxu98-grobid6.hf.space", 
    "https://qingxu98-grobid7.hf.space", "https://qingxu98-grobid8.hf.space", 
]


# Allow modifying the configuration of this page through natural language description，This feature has a certain level of danger，Default closed
ALLOW_RESET_CONFIG = False


# When using the AutoGen plugin，Whether to run the code using Docker container
AUTOGEN_USE_DOCKER = False


# Temporary upload folder location，Please do not modify
PATH_PRIVATE_UPLOAD = "private_upload"


# Location of the log folder，Please do not modify
PATH_LOGGING = "gpt_log"


# In addition to connecting OpenAI，What other occasions allow the use of proxies，Please do not modify
WHEN_TO_USE_PROXY = ["Download_LLM", "Download_Gradio_Theme", "Connect_Grobid", 
                     "Warmup_Modules", "Nougat_Download", "AutoGen"]


# *Experimental feature*: Automatically detect and block invalid keys，Please do not use
BLOCK_INVALID_APIKEY = False


# Maximum number limit for custom buttons
NUM_CUSTOM_BASIC_BTN = 4

"""
Online large model configuration relationship diagram
│
├── "gpt-3.5-turbo" 等openai模型
│   ├── API_KEY
│   ├── CUSTOM_API_KEY_PATTERN（Not commonly used）
│   ├── API_ORG（Not commonly used）
│   └── API_URL_REDIRECT（Not commonly used）
│
├── "azure-gpt-3.5" 等azure模型（Single Azure model，No need for dynamic switching）
│   ├── API_KEY
│   ├── AZURE_ENDPOINT
│   ├── AZURE_API_KEY
│   ├── AZURE_ENGINE
│   └── API_URL_REDIRECT
│
├── "azure-gpt-3.5" 等azure模型（Multiple Azure models，Dynamic switching is required，High priority）
│   └── AZURE_CFG_ARRAY
│
├── "spark" Spark Cognitive Big Model spark & sparkv2
│   ├── XFYUN_APPID
│   ├── XFYUN_API_SECRET
│   └── XFYUN_API_KEY
│
├── "claude-1-100k" 等claude模型
│   └── ANTHROPIC_API_KEY
│
├── "stack-claude"
│   ├── SLACK_CLAUDE_BOT_ID
│   └── SLACK_CLAUDE_USER_TOKEN
│
├── "qianfan" Baidu Qianfan大模型库
│   ├── BAIDU_CLOUD_QIANFAN_MODEL
│   ├── BAIDU_CLOUD_API_KEY
│   └── BAIDU_CLOUD_SECRET_KEY
│
├── "newbing" Newbing接口不再稳定，Not recommended for use
    ├── NEWBING_STYLE
    └── NEWBING_COOKIES

    
Diagram of user interface layout dependencies
│
├── CHATBOT_HEIGHT Height of the chat window
├── CODE_HIGHLIGHT Code highlighting
├── LAYOUT window layout
├── DARK_MODE Dark mode / Light mode
├── DEFAULT_FN_GROUPS Plugin classification default options
├── THEME color theme
├── AUTO_CLEAR_TXT Whether to automatically clear the input box when submitting
├── ADD_WAIFU Add a live2d decoration
├── ALLOW_RESET_CONFIG Whether to allow modifying the configuration of this page through natural language description，This feature has a certain level of danger


Plugin online service configuration dependency diagram
│
├── Voice function
│   ├── ENABLE_AUDIO
│   ├── ALIYUN_TOKEN
│   ├── ALIYUN_APPKEY
│   ├── ALIYUN_ACCESSKEY
│   └── ALIYUN_SECRET
│
├── Accurate parsing of PDF documents
│   └── GROBID_URLS

"""



import time
import threading
from void_terminal.toolbox import update_ui, Singleton
from multiprocessing import Process, Pipe
from contextlib import redirect_stdout
from void_terminal.request_llms.queued_pipe import create_queue_pipe

class ThreadLock(object):
    def __init__(self):
        self._lock = threading.Lock()

    def acquire(self):
        # print("acquiring", self)
        #traceback.print_tb
        self._lock.acquire()
        # print("acquired", self)

    def release(self):
        # print("released", self)
        #traceback.print_tb
        self._lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, traceback):
        self.release()

@Singleton
class GetSingletonHandle():
    def __init__(self):
        self.llm_model_already_running = {}

    def get_llm_model_instance(self, cls, *args, **kargs):
        if cls not in self.llm_model_already_running:
            self.llm_model_already_running[cls] = cls(*args, **kargs)
            return self.llm_model_already_running[cls]
        elif self.llm_model_already_running[cls].corrupted:
            self.llm_model_already_running[cls] = cls(*args, **kargs)
            return self.llm_model_already_running[cls]
        else:
            return self.llm_model_already_running[cls]

def reset_tqdm_output():
    import sys, tqdm
    def status_printer(self, file):
        fp = file
        if fp in (sys.stderr, sys.stdout):
            getattr(sys.stderr, 'flush', lambda: None)()
            getattr(sys.stdout, 'flush', lambda: None)()

        def fp_write(s):
            print(s)
        last_len = [0]

        def print_status(s):
            from tqdm.utils import disp_len
            len_s = disp_len(s)
            fp_write('\r' + s + (' ' * max(last_len[0] - len_s, 0)))
            last_len[0] = len_s
        return print_status
    tqdm.tqdm.status_printer = status_printer


class LocalLLMHandle(Process):
    def __init__(self):
        # ⭐run in main process
        super().__init__(daemon=True)
        self.is_main_process = True # init
        self.corrupted = False
        self.load_model_info()
        self.parent, self.child = create_queue_pipe()
        self.parent_state, self.child_state = create_queue_pipe()
        # allow redirect_stdout
        self.std_tag = "[Subprocess Message] "
        self.running = True
        self._model = None
        self._tokenizer = None
        self.state = ""
        self.check_dependency()
        self.is_main_process = False    # state wrap for child process
        self.start()
        self.is_main_process = True     # state wrap for child process
        self.threadLock = ThreadLock()

    def get_state(self):
        # ⭐run in main process
        while self.parent_state.poll():
            self.state = self.parent_state.recv()
        return self.state

    def set_state(self, new_state):
        # ⭐run in main process or 🏃‍♂️🏃‍♂️🏃‍♂️ run in child process 
        if self.is_main_process:
            self.state = new_state
        else:
            self.child_state.send(new_state)

    def load_model_info(self):
        # 🏃‍♂️🏃‍♂️🏃‍♂️ run in child process
        raise NotImplementedError("Method not implemented yet")
        self.model_name = ""
        self.cmd_to_install = ""

    def load_model_and_tokenizer(self):
        """
        This function should return the model and the tokenizer
        """
        # 🏃‍♂️🏃‍♂️🏃‍♂️ run in child process
        raise NotImplementedError("Method not implemented yet")

    def llm_stream_generator(self, **kwargs):
        # 🏃‍♂️🏃‍♂️🏃‍♂️ run in child process
        raise NotImplementedError("Method not implemented yet")

    def try_to_import_special_deps(self, **kwargs):
        """
        import something that will raise error if the user does not install requirement_*.txt
        """
        # ⭐run in main process
        raise NotImplementedError("Method not implemented yet")

    def check_dependency(self):
        # ⭐run in main process
        try:
            self.try_to_import_special_deps()
            self.set_state("`Dependency check passed`")
            self.running = True
        except:
            self.set_state(f"Missing{self.model_name}Dependencies of，If you want to use{self.model_name}，In addition to the basic pip dependencies，You also need to run{self.cmd_to_install}Installation{self.model_name}Dependencies of。")
            self.running = False

    def run(self):
        # 🏃‍♂️🏃‍♂️🏃‍♂️ run in child process
        # First run，Load parameters
        self.child.flush = lambda *args: None
        self.child.write = lambda x: self.child.send(self.std_tag + x)
        reset_tqdm_output()
        self.set_state("`Try to load the model`")
        try:
            with redirect_stdout(self.child):
                self._model, self._tokenizer = self.load_model_and_tokenizer()
        except:
            self.set_state("`Failed to load the model`")
            self.running = False
            from void_terminal.toolbox import trimmed_format_exc
            self.child.send(
                f'[Local Message] Cannot load properly{self.model_name}Parameters of.' + '\n```\n' + trimmed_format_exc() + '\n```\n')
            self.child.send('[FinishBad]')
            raise RuntimeError(f"Cannot load properly{self.model_name}parameters!")

        self.set_state("`Ready`")
        while True:
            # Enter task waiting state
            kwargs = self.child.recv()
            # Received message，Start requesting
            try:
                for response_full in self.llm_stream_generator(**kwargs):
                    self.child.send(response_full)
                    # print('debug' + response_full)
                self.child.send('[Finish]')
                # Request processing ends，Start the next loop
            except:
                from void_terminal.toolbox import trimmed_format_exc
                self.child.send(
                    f'[Local Message] Call{self.model_name}Failure.' + '\n```\n' + trimmed_format_exc() + '\n```\n')
                self.child.send('[Finish]')

    def clear_pending_messages(self):
        # ⭐run in main process
        while True:
            if  self.parent.poll():
                self.parent.recv()
                continue
            for _ in range(5):
                time.sleep(0.5)
                if  self.parent.poll():
                    r = self.parent.recv()
                    continue
            break
        return 
    
    def stream_chat(self, **kwargs):
        # ⭐run in main process
        if self.get_state() == "`Ready`":
            yield "`Waiting for thread lock，Please wait in line ...`"

        with self.threadLock:
            if self.parent.poll():
                yield "`Please wait in line ...`"
                self.clear_pending_messages()
            self.parent.send(kwargs)
            std_out = ""
            std_out_clip_len = 4096
            while True:
                res = self.parent.recv()
                # pipe_watch_dog.feed()
                if res.startswith(self.std_tag):
                    new_output = res[len(self.std_tag):]
                    std_out = std_out[:std_out_clip_len]
                    # print(new_output, end='')
                    std_out = new_output + std_out
                    yield self.std_tag + '\n```\n' + std_out + '\n```\n'
                elif res == '[Finish]':
                    break
                elif res == '[FinishBad]':
                    self.running = False
                    self.corrupted = True
                    break
                else:
                    std_out = ""
                    yield res

def get_local_llm_predict_fns(LLMSingletonClass, model_name, history_format='classic'):
    load_message = f"{model_name}Not loaded yet，Loading takes some time。Attention，Depending on`config.py`Configuration，{model_name}Consumes a large amount of memory（CPU）Or video memory（GPU），May cause low-end computers to freeze..."

    def predict_no_ui_long_connection(inputs, llm_kwargs, history=[], sys_prompt="", observe_window=[], console_slience=False):
        """
            refer to request_llms/bridge_all.py
        """
        _llm_handle = GetSingletonHandle().get_llm_model_instance(LLMSingletonClass)
        if len(observe_window) >= 1:
            observe_window[0] = load_message + "\n\n" + _llm_handle.get_state()
        if not _llm_handle.running:
            raise RuntimeError(_llm_handle.get_state())

        if history_format == 'classic':
            # No sys_prompt interface，Therefore, add prompt to history
            history_feedin = []
            history_feedin.append([sys_prompt, "Certainly!"])
            for i in range(len(history)//2):
                history_feedin.append([history[2*i], history[2*i+1]])
        elif history_format == 'chatglm3':
            # There is a sys_prompt interface
            conversation_cnt = len(history) // 2
            history_feedin = [{"role": "system", "content": sys_prompt}]
            if conversation_cnt:
                for index in range(0, 2*conversation_cnt, 2):
                    what_i_have_asked = {}
                    what_i_have_asked["role"] = "user"
                    what_i_have_asked["content"] = history[index]
                    what_gpt_answer = {}
                    what_gpt_answer["role"] = "assistant"
                    what_gpt_answer["content"] = history[index+1]
                    if what_i_have_asked["content"] != "":
                        if what_gpt_answer["content"] == "":
                            continue
                        history_feedin.append(what_i_have_asked)
                        history_feedin.append(what_gpt_answer)
                    else:
                        history_feedin[-1]['content'] = what_gpt_answer['content']

        watch_dog_patience = 5  # Watchdog (watchdog) Patience, Set 5 seconds
        response = ""
        for response in _llm_handle.stream_chat(query=inputs, history=history_feedin, max_length=llm_kwargs['max_length'], top_p=llm_kwargs['top_p'], temperature=llm_kwargs['temperature']):
            if len(observe_window) >= 1:
                observe_window[0] = response
            if len(observe_window) >= 2:
                if (time.time()-observe_window[1]) > watch_dog_patience:
                    raise RuntimeError("Program terminated。")
        return response

    def predict(inputs, llm_kwargs, plugin_kwargs, chatbot, history=[], system_prompt='', stream=True, additional_fn=None):
        """
            refer to request_llms/bridge_all.py
        """
        chatbot.append((inputs, ""))

        _llm_handle = GetSingletonHandle().get_llm_model_instance(LLMSingletonClass)
        chatbot[-1] = (inputs, load_message + "\n\n" + _llm_handle.get_state())
        yield from update_ui(chatbot=chatbot, history=[])
        if not _llm_handle.running:
            raise RuntimeError(_llm_handle.get_state())

        if additional_fn is not None:
            from void_terminal.core_functional import handle_core_functionality
            inputs, history = handle_core_functionality(
                additional_fn, inputs, history, chatbot)

        # Process historical information
        if history_format == 'classic':
            # No sys_prompt interface，Therefore, add prompt to history
            history_feedin = []
            history_feedin.append([system_prompt, "Certainly!"])
            for i in range(len(history)//2):
                history_feedin.append([history[2*i], history[2*i+1]])
        elif history_format == 'chatglm3':
            # There is a sys_prompt interface
            conversation_cnt = len(history) // 2
            history_feedin = [{"role": "system", "content": system_prompt}]
            if conversation_cnt:
                for index in range(0, 2*conversation_cnt, 2):
                    what_i_have_asked = {}
                    what_i_have_asked["role"] = "user"
                    what_i_have_asked["content"] = history[index]
                    what_gpt_answer = {}
                    what_gpt_answer["role"] = "assistant"
                    what_gpt_answer["content"] = history[index+1]
                    if what_i_have_asked["content"] != "":
                        if what_gpt_answer["content"] == "":
                            continue
                        history_feedin.append(what_i_have_asked)
                        history_feedin.append(what_gpt_answer)
                    else:
                        history_feedin[-1]['content'] = what_gpt_answer['content']

        # Start receiving replies
        response = f"[Local Message] Waiting{model_name}Responding ..."
        for response in _llm_handle.stream_chat(query=inputs, history=history_feedin, max_length=llm_kwargs['max_length'], top_p=llm_kwargs['top_p'], temperature=llm_kwargs['temperature']):
            chatbot[-1] = (inputs, response)
            yield from update_ui(chatbot=chatbot, history=history)

        # Summary output
        if response == f"[Local Message] Waiting{model_name}Responding ...":
            response = f"[Local Message] {model_name}Response exception ..."
        history.extend([inputs, response])
        yield from update_ui(chatbot=chatbot, history=history)

    return predict_no_ui_long_connection, predict



import numpy as np
from scipy import interpolate

def Singleton(cls):
    _instance = {}
 
    def _singleton(*args, **kargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kargs)
        return _instance[cls]
 
    return _singleton


@Singleton
class RealtimeAudioDistribution():
    def __init__(self) -> None:
        self.data = {}
        self.max_len = 1024*1024
        self.rate = 48000   # Read-only，Number of samples per second

    def clean_up(self):
        self.data = {}

    def feed(self, uuid, audio):
        self.rate, audio_ = audio
        # print('feed', len(audio_), audio_[-25:])
        if uuid not in self.data:
            self.data[uuid] = audio_
        else:
            new_arr = np.concatenate((self.data[uuid], audio_))
            if len(new_arr) > self.max_len: new_arr = new_arr[-self.max_len:]
            self.data[uuid] = new_arr

    def read(self, uuid):
        if uuid in self.data:
            res = self.data.pop(uuid)
            # print('\r read-', len(res), '-', max(res), end='', flush=True)
        else:
            res = None
        return res
    
def change_sample_rate(audio, old_sr, new_sr):
    duration = audio.shape[0] / old_sr

    time_old  = np.linspace(0, duration, audio.shape[0])
    time_new  = np.linspace(0, duration, int(audio.shape[0] * new_sr / old_sr))

    interpolator = interpolate.interp1d(time_old, audio.T)
    new_audio = interpolator(time_new).T
    return new_audio.astype(np.int16)


from void_terminal.toolbox import HotReload  # HotReload means hot update，After modifying the function plugin，No need to restart the program，The code takes effect directly


def get_crazy_functions():
    from void_terminal.crazy_functions.ReadArticleWriteSummary import ReadArticleWriteSummary
    from void_terminal.crazy_functions.GenerateFunctionComments import BatchGenerateFunctionComments
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseProjectItself
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParsePythonProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import AnalyzeAMatlabProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseCProjectHeaderFiles
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseCProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseGolangProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseRustProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseJavaProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParseFrontendProject
    from void_terminal.crazy_functions.AdvancedFunctionTemplate import HighOrderFunctionTemplateFunctions
    from void_terminal.crazy_functions.FullTextProofreadingForLatex import EnglishProofreadingForLatex
    from void_terminal.crazy_functions.InquiryMultipleLargeLanguageModels import SimultaneousInquiry
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParsingLuaProject
    from void_terminal.crazy_functions.ParseProjectSourceCode import ParsingCSharpProject
    from void_terminal.crazy_functions.SummarizingWordDocuments import SummarizingWordDocuments
    from void_terminal.crazy_functions.ParsingJupyterNotebook import ParsingIpynbFiles
    from void_terminal.crazy_functions.ConversationHistoryArchive import ConversationHistoryArchive
    from void_terminal.crazy_functions.ConversationHistoryArchive import LoadConversationHistoryArchive
    from void_terminal.crazy_functions.ConversationHistoryArchive import DeleteAllLocalConversationHistoryRecords
    from void_terminal.crazy_functions.Accessibility import ClearCache
    from void_terminal.crazy_functions.BatchTranslateMarkdown import TranslateMarkdownFromEnglishToChinese
    from void_terminal.crazy_functions.BatchSummarizePDFDocuments import BatchSummarizePDFDocuments
    from void_terminal.crazy_functions.BatchTranslatePDFDocuments_MultiThreaded import BatchTranslatePDFDocuments
    from void_terminal.crazy_functions.GoogleSearchAssistant import GoogleSearchAssistant
    from void_terminal.crazy_functions.UnderstandPdfDocumentContent import UnderstandPdfDocumentContentStandardFileInput
    from void_terminal.crazy_functions.FullTextProofreadingForLatex import LatexChineseProofreading
    from void_terminal.crazy_functions.FullTextProofreadingForLatex import LatexEnglishCorrection
    from void_terminal.crazy_functions.LatexFullTextTranslation import LatexChineseToEnglish
    from void_terminal.crazy_functions.LatexFullTextTranslation import LatexEnglishToChinese
    from void_terminal.crazy_functions.BatchTranslateMarkdown import MarkdownChineseToEnglish
    from void_terminal.crazy_functions.VoidTerminal import VoidTerminal


    function_plugins = {
        "VoidTerminal": {
            "Group": "Dialogue|Programming|Academic|Intelligent agent",
            "Color": "stop",
            "AsButton": True,
            "Function": HotReload(VoidTerminal)
        },
        "Parse the entire Python project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": True,
            "Info": "All source files of ParsePythonProject(.py) | Input parameter is the path",
            "Function": HotReload(ParsePythonProject)
        },
        "LoadConversationHistoryArchive（Upload archive or enter path first）": {
            "Group": "Conversation",
            "Color": "stop",
            "AsButton": False,
            "Info": "LoadConversationHistoryArchive | Input parameter is a path",
            "Function": HotReload(LoadConversationHistoryArchive)
        },
        "DeleteAllLocalConversationHistoryRecords（Handle with caution）": {
            "Group": "Conversation",
            "AsButton": False,
            "Info": "DeleteAllLocalConversationHistoryRecords，Handle with caution | No input parameters required",
            "Function": HotReload(DeleteAllLocalConversationHistoryRecords)
        },
        "Clear all cache files（Handle with caution）": {
            "Group": "Conversation",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Clear all cache files，Handle with caution | No input parameters required",
            "Function": HotReload(ClearCache)
        },
        "Batch summarize Word documents": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,
            "Info": "Batch SummarizingWordDocuments | Input parameter is the path",
            "Function": HotReload(SummarizingWordDocuments)
        },
        "Parse the entire Matlab project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,
            "Info": "All source files of AnalyzeAMatlabProject(.m) | Input parameter is the path",
            "Function": HotReload(AnalyzeAMatlabProject)
        },
        "Parse the entire C++ project header file": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Parse all header files of a C++ project(.h/.hpp) | Input parameter is the path",
            "Function": HotReload(ParseCProjectHeaderFiles)
        },
        "Parse the entire C++ project（.cpp/.hpp/.c/.h）": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Parse all source files of a C++ project（.cpp/.hpp/.c/.h）| Input parameter is the path",
            "Function": HotReload(ParseCProject)
        },
        "Parse the entire Go project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Parse all source files of a Go project | Input parameter is path",
            "Function": HotReload(ParseGolangProject)
        },
        "Parse the entire Go project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "All source files of ParseRustProject | Input parameter is path",
            "Function": HotReload(ParseRustProject)
        },
        "Parse the entire Java project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "All source files of ParseJavaProject | Input parameter is path",
            "Function": HotReload(ParseJavaProject)
        },
        "Parse the entire front-end project（js,ts,CSS, etc.）": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Parse all source files of ParseFrontendProject（js,ts,CSS, etc.） | Input parameter is the path",
            "Function": HotReload(ParseFrontendProject)
        },
        "Parse the entire Lua project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "All source files of ParsingLuaProject | Input parameter is path",
            "Function": HotReload(ParsingLuaProject)
        },
        "Parse the entire C# project": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "ParsingCSharpProject`s all source files | Input parameter is a path",
            "Function": HotReload(ParsingCSharpProject)
        },
        "Parse Jupyter Notebook files": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,
            "Info": "Parse Jupyter Notebook file | Input parameter is path",
            "Function": HotReload(ParsingIpynbFiles),
            "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
            "ArgsReminder": "If 0 is entered，Do not parse Markdown blocks in the notebook",  # Display prompt in the advanced parameter input area
        },
        "Read Tex paper and write abstract": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,
            "Info": "Read Tex paper and write abstract | Input parameter is the path",
            "Function": HotReload(ReadArticleWriteSummary)
        },
        "Translate README or MD": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": True,
            "Info": "Translate Markdown to Chinese | Input parameters are path or URL",
            "Function": HotReload(TranslateMarkdownFromEnglishToChinese)
        },
        "Translate Markdown or README（Support Github links）": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,
            "Info": "Translate Markdown or README to Chinese | Input parameters are path or URL",
            "Function": HotReload(TranslateMarkdownFromEnglishToChinese)
        },
        "BatchGenerateFunctionComments": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Batch generate function comments | Input parameter is the path",
            "Function": HotReload(BatchGenerateFunctionComments)
        },
        "Accurate translation of PDF paper": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": True,  
            "Info": "Translate PDF papers accurately into Chinese | Input parameter is the path",
            "Function": HotReload(BatchTranslatePDFDocuments)
        },
        "Save the current conversation": {
            "Group": "Conversation",
            "AsButton": True,
            "Info": "Save current conversation | No input parameters required",
            "Function": HotReload(ConversationHistoryArchive)
        },
        "[Multithreading demo]Parse this project itself（Translate the source code）": {
            "Group": "Conversation|Programming",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Parse and translate the source code of this project in multi-threading | No input parameters required",
            "Function": HotReload(ParseProjectItself)
        },
        "Today in history": {
            "Group": "Conversation",
            "AsButton": True,
            "Info": "View events from history (This is a plugin demo for developers) | No input parameters required",
            "Function": HotReload(HighOrderFunctionTemplateFunctions)
        },
        "Inquire multiple GPT models": {
            "Group": "Conversation",
            "Color": "stop",
            "AsButton": True,
            "Function": HotReload(SimultaneousInquiry)
        },
        "BatchSummarizePDFDocuments": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Content of BatchSummarizePDFDocuments | Input parameter is a path",
            "Function": HotReload(BatchSummarizePDFDocuments)
        },
        "Google Scholar search assistant（Enter the URL of Google Scholar search page）": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Use Google Scholar search assistant to search for results of a specific URL | Input parameter is the URL of Google Scholar search page",
            "Function": HotReload(GoogleSearchAssistant)
        },
        "UnderstandPdfDocumentContent （Imitate ChatPDF）": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Understand the content of the PDF document and answer | Input parameter is path",
            "Function": HotReload(UnderstandPdfDocumentContentStandardFileInput)
        },
        "English Latex project full text proofreading（Input path or upload compressed package）": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Polish the full text of English Latex projects | Input parameters are paths or uploaded compressed packages",
            "Function": HotReload(EnglishProofreadingForLatex)
        },
        "Full-text correction of English Latex projects（Input path or upload compressed package）": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Correct the entire English Latex project | Input parameter is the path or upload compressed package",
            "Function": HotReload(LatexEnglishCorrection)
        },
        "Chinese Latex project full text proofreading（Input path or upload compressed package）": {
            "Group": "Academic",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Polish the entire text of a Chinese Latex project | Input parameter is path or upload compressed package",
            "Function": HotReload(LatexChineseProofreading)
        },

        # Has been replaced by a new plugin
        # "Latex project full text translation from Chinese to English（Input path or upload compressed package）": {
        #     "Group": "Academic",
        #     "Color": "stop",
        #     "AsButton": False,  # Add to the drop-down menu
        #     "Info": "Translate the full text of Latex projects from Chinese to English | Input parameter is the path or upload a compressed package",
        #     "Function": HotReload(LatexChineseToEnglish)
        # },

        # Has been replaced by a new plugin
        # "Latex project full text translation from English to Chinese（Input path or upload compressed package）": {
        #     "Group": "Academic",
        #     "Color": "stop",
        #     "AsButton": False,  # Add to the drop-down menu
        #     "Info": "Translate the entire text of Latex project from English to Chinese | Input parameters are path or uploaded compressed package",
        #     "Function": HotReload(LatexEnglishToChinese)
        # },
        
        "Batch Markdown Chinese to English（Input path or upload compressed package）": {
            "Group": "programming",
            "Color": "stop",
            "AsButton": False,  # Add to the drop-down menu
            "Info": "Batch translate Chinese to English in Markdown files | Input parameter is a path or upload a compressed package",
            "Function": HotReload(MarkdownChineseToEnglish)
        },
    }

    # -=--=- Experimental plugins that have not been fully tested & plugins that require additional dependencies -=--=-
    try:
        from void_terminal.crazy_functions.DownloadArxivPaperTranslateAbstract import DownloadArxivPaperAndTranslateAbstract
        function_plugins.update({
            "One-click Download Arxiv Paper and Translate Abstract（Enter the number in input first，e.g. 1812.10695）": {
                "Group": "Academic",
                "Color": "stop",
                "AsButton": False,  # Add to the drop-down menu
                # "Info": "DownloadArxivPaperAndTranslateAbstract | InputParameters为arxiv编号e.g. 1812.10695",
                "Function": HotReload(DownloadArxivPaperAndTranslateAbstract)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.ChatGPTConnectedToNetwork import ConnectToNetworkToAnswerQuestions
        function_plugins.update({
            "ConnectToNetworkToAnswerQuestions（Click the plugin after entering the question，Access to Google is required）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,  # Add to the drop-down menu
                # "Info": "ConnectToNetworkToAnswerQuestions（Access to Google is required）| Input parameter is a question",
                "Function": HotReload(ConnectToNetworkToAnswerQuestions)
            }
        })
        from void_terminal.crazy_functions.OnlineChatGPT_BingEdition import ConnectBingSearchAnswerQuestion
        function_plugins.update({
            "ConnectToNetworkToAnswerQuestions（Chinese Bing version，Click the plugin after entering the question）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,  # Add to the drop-down menu
                "Info": "ConnectToNetworkToAnswerQuestions（Need to access Chinese Bing）| Input parameter is a question",
                "Function": HotReload(ConnectBingSearchAnswerQuestion)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.ParseProjectSourceCode import ParseAnyCodeProject
        function_plugins.update({
            "ParseProjectSourceCode（Manually specify and filter the source code file type）": {
                "Group": "programming",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
                "ArgsReminder": "Separate with commas when entering, * stands for wildcard, Adding ^ means not matching; Not entering means matching all。For example: \"*.c, ^*.cpp, config.toml, ^*.toml\"",  # Display prompt in the advanced parameter input area
                "Function": HotReload(ParseAnyCodeProject)
            },
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.InquiryMultipleLargeLanguageModels import InquireSimultaneously_SpecifiedModel
        function_plugins.update({
            "Inquire multiple GPT models（Manually specify which models to ask）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
                "ArgsReminder": "Support any number of llm interfaces，Separate with & symbol。e.g. chatglm&gpt-3.5-turbo&api2d-gpt-4",  # Display prompt in the advanced parameter input area
                "Function": HotReload(InquireSimultaneously_SpecifiedModel)
            },
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.ImageGeneration import ImageGeneration_DALLE2, ImageGeneration_DALLE3, ImageModification_DALLE2
        function_plugins.update({
            "ImageGeneration_DALLE2（Switch the model to openai or api2d first）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
                "ArgsReminder": "Enter the resolution here, e.g. 1024x1024（Default），Supports 256x256, 512x512, 1024x1024",  # Display prompt in the advanced parameter input area
                "Info": "Generate images using DALLE2 | Input parameter string，Provide the content of the image",
                "Function": HotReload(ImageGeneration_DALLE2)
            },
        })
        function_plugins.update({
            "ImageGeneration_DALLE3（Switch the model to openai or api2d first）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
                "ArgsReminder": "Enter the resolution here, e.g. 1024x1024（Default），Support 1024x1024, 1792x1024, 1024x1792",  # Display prompt in the advanced parameter input area
                "Info": "Use DALLE3 to generate images | Input parameter string，Provide the content of the image",
                "Function": HotReload(ImageGeneration_DALLE3)
            },
        })
        # function_plugins.update({
        #     "ImageModification_DALLE2（Launching DALLE2 image modification wizard program）": {
        #         "Group": "Conversation",
        #         "Color": "stop",
        #         "AsButton": False,
        #         "AdvancedArgs": True,  # When calling，Invoke the advanced parameter input area（Default is False）
        #         "ArgsReminder": "Enter the resolution here, e.g. 1024x1024（Default），Support 1024x1024, 1792x1024, 1024x1792",  # Display prompt in the advanced parameter input area
        #         # "Info": "UseDALLE2修改Image | InputParametersString，Provide the content of the image",
        #         "Function": HotReload(ImageModification_DALLE2)
        #     },
        # })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.SummaryAudioVideo import SummaryAudioVideo
        function_plugins.update({
            "Batch Summary Audio Video（Input path or upload compressed package）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder": "Call openai api to use whisper-1 model, Supported formats at present:mp4, m4a, wav, mpga, mpeg, mp3。Parsing tips can be entered here，For example：Parse to Simplified Chinese（Default）。",
                "Info": "Batch summarize audio or video | Input parameter is path",
                "Function": HotReload(SummaryAudioVideo)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.MathematicalAnimationGenerationManim import AnimationGeneration
        function_plugins.update({
            "Mathematical Animation Generation（Manim）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "Info": "Generate an animation based on natural language description | Input parameter is a sentence",
                "Function": HotReload(AnimationGeneration)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.BatchTranslateMarkdown import TranslateMarkdownToSpecifiedLanguage
        function_plugins.update({
            "Markdown translation（Specify the language to translate into）": {
                "Group": "programming",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder": "Please enter which language to translate into，Default is Chinese。",
                "Function": HotReload(TranslateMarkdownToSpecifiedLanguage)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.LangchainKnowledgeBase import UpdateKnowledgeArchive
        function_plugins.update({
            "Building knowledge base（Upload file materials first,Run this plugin again）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder": "The knowledge base name ID to be injected here, Default is `default`。Files can be saved for a long time after entering the knowledge base。You can use this plugin again by calling it，Append more documents to the knowledge base。",
                "Function": HotReload(UpdateKnowledgeArchive)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.LangchainKnowledgeBase import ReadKnowledgeArchiveAnswerQuestions
        function_plugins.update({
            "UpdateKnowledgeArchive（After building the knowledge base,Run this plugin again）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder": "Knowledge base name ID to be extracted, Default is `default`, You need to build the knowledge base before running this plugin。",
                "Function": HotReload(ReadKnowledgeArchiveAnswerQuestions)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.InteractiveFunctionFunctionTemplate import InteractiveFunctionTemplateFunction
        function_plugins.update({
            "Interactive function template Demo function（Search for wallpapers on wallhaven.cc）": {
                "Group": "Conversation",
                "Color": "stop",
                "AsButton": False,
                "Function": HotReload(InteractiveFunctionTemplateFunction)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.OutputPDFFromLatex import CorrectEnglishInLatexWithPDFComparison
        function_plugins.update({
            "Latex English correction + highlight correction position [Requires Latex]": {
                "Group": "Academic",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder": "If necessary, Please append more detailed correction instructions here（Use English）。",
                "Function": HotReload(CorrectEnglishInLatexWithPDFComparison)
            }
        })
        from void_terminal.crazy_functions.OutputPDFFromLatex import TranslateChineseToEnglishInLatexAndRecompilePDF
        function_plugins.update({
            "Fine translation of Arixv paper（Enter arxivID）[Requires Latex]": {
                "Group": "Academic",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder":
                    "If necessary, Please provide custom translation command here, Resolve the issue of inaccurate translation for some terms。 " +
                    "For example当单词'agent'Translation不准确When, Please try copying the following instructions to the advanced parameters section: " +
                    'If the term "agent" is used in this section, it should be translated to "Intelligent agent". ',
                "Info": "Fine translation of Arixv paper | Input parameter is the ID of arxiv paper，For example, 1812.10695",
                "Function": HotReload(TranslateChineseToEnglishInLatexAndRecompilePDF)
            }
        })
        function_plugins.update({
            "Local Latex paper fine translation（Upload Latex project）[Requires Latex]": {
                "Group": "Academic",
                "Color": "stop",
                "AsButton": False,
                "AdvancedArgs": True,
                "ArgsReminder":
                    "If necessary, Please provide custom translation command here, Resolve the issue of inaccurate translation for some terms。 " +
                    "For example当单词'agent'Translation不准确When, Please try copying the following instructions to the advanced parameters section: " +
                    'If the term "agent" is used in this section, it should be translated to "Intelligent agent". ',
                "Info": "Locally translate Latex papers with fine-grained translation | Input parameter is the path",
                "Function": HotReload(TranslateChineseToEnglishInLatexAndRecompilePDF)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.toolbox import get_conf
        ENABLE_AUDIO = get_conf('ENABLE_AUDIO')
        if ENABLE_AUDIO:
            from void_terminal.crazy_functions.VoiceAssistant import VoiceAssistant
            function_plugins.update({
                "Real-time voice conversation": {
                    "Group": "Conversation",
                    "Color": "stop",
                    "AsButton": True,
                    "Info": "This is a voice conversation assistant that is always listening | No input parameters",
                    "Function": HotReload(VoiceAssistant)
                }
            })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.BatchTranslatePDFDocuments_NOUGAT import BatchTranslatePDFDocuments
        function_plugins.update({
            "Accurate translation of PDF documents（NOUGAT）": {
                "Group": "Academic",
                "Color": "stop",
                "AsButton": False,
                "Function": HotReload(BatchTranslatePDFDocuments)
            }
        })
    except:
        print('Load function plugin failed')

    try:
        from void_terminal.crazy_functions.DynamicFunctionGeneration import DynamicFunctionGeneration
        function_plugins.update({
            "Dynamic code interpreter（CodeInterpreter）": {
                "Group": "Intelligent agent",
                "Color": "stop",
                "AsButton": False,
                "Function": HotReload(DynamicFunctionGeneration)
            }
        })
    except:
        print('Load function plugin failed')

    from void_terminal.crazy_functions.MultiAgent import MultiAgentTerminal
    function_plugins.update({
        "AutoGenMultiAgentTerminal": {
            "Group": "Intelligent agent",
            "Color": "stop",
            "AsButton": True,
            "Function": HotReload(MultiAgentTerminal)
        }
    })

    # try:
    #     from crazy_functions.chatglm fine-tuning tool import fine-tuning dataset generation
    #     function_plugins.update({
    #         "黑盒模型学习: FineTuneDatasetGeneration (Upload dataset first)": {
    #             "Color": "stop",
    #             "AsButton": False,
    #             "AdvancedArgs": True,
    #             "ArgsReminder": "针对数据集Input（E.g. green hat * dark blue shirt * black sports pants）Give instructions，For example, you can copy the following command below: --llm_to_learn=azure-gpt-3.5 --prompt_prefix='根据下面的服装类型prompt，Imagine a wearer，Describe the appearance, environment, inner world, and past experiences of this person。Requirement：Within 100 words，Use the second person。' --system_prompt=''",
    #             "Function": HotReload(FineTuneDatasetGeneration)
    #         }
    #     })
    # except:
    #     print('Load function plugin failed')



    """
    Set default value:
    - Default Group = Conversation
    - Default AsButton = True
    - Default AdvancedArgs = False
    - Default Color = secondary
    """
    for name, function_meta in function_plugins.items():
        if "Group" not in function_meta:
            function_plugins[name]["Group"] = 'Conversation'
        if "AsButton" not in function_meta:
            function_plugins[name]["AsButton"] = True
        if "AdvancedArgs" not in function_meta:
            function_plugins[name]["AdvancedArgs"] = False
        if "Color" not in function_meta:
            function_plugins[name]["Color"] = 'secondary'

    return function_plugins



from void_terminal.toolbox import CatchException, update_ui
from void_terminal.crazy_functions.crazy_utils import request_gpt_model_in_new_thread_with_ui_alive


@CatchException
def InteractiveFunctionTemplateFunction(txt, llm_kwargs, plugin_kwargs, chatbot, history, system_prompt, web_port):
    """
    txt             Text entered by the user in the input field，For example, a paragraph that needs to be translated，For example, a file path that contains files to be processed
    llm_kwargs      GPT model parameters, Such as temperature and top_p, Generally pass it on as is
    plugin_kwargs   Plugin model parameters, Such as temperature and top_p, Generally pass it on as is
    chatbot         Chat display box handle，Displayed to the user
    history         Chat history，Context summary
    system_prompt   Silent reminder to GPT
    web_port        Current software running port number
    """
    history = []    # Clear history，To avoid input overflow
    chatbot.append(("What is this function？", "InteractiveFunctionFunctionTemplate。After execution is complete, You can store your own status in cookies, Waiting for the user to call again。"))
    yield from update_ui(chatbot=chatbot, history=history) # Refresh the page

    state = chatbot._cookies.get('plugin_state_0001', None) # Initializing plugin status

    if state is None:
        chatbot._cookies['lock_plugin'] = 'crazy_functions.InteractiveFunctionFunctionTemplate->InteractiveFunctionTemplateFunction'      # Assign plugin lock, lock plugin callback path，When the next user submits，Will directly jump to the function
        chatbot._cookies['plugin_state_0001'] = 'wait_user_keyword'                              # Assign plugin status

        chatbot.append(("First call：", "Please enter a keyword., I will search for related wallpapers for you, Suggest using English words, Plugin is locked，Submit directly。"))
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return

    if state == 'wait_user_keyword':
        chatbot._cookies['lock_plugin'] = None          # Unlock plugin，Avoid forgetting to cause deadlock
        chatbot._cookies['plugin_state_0001'] = None    # Release plugin status，Avoid forgetting to cause deadlock

        # Unlock plugin
        chatbot.append((f"Get keywords：{txt}", ""))
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        page_return = get_image_page_by_keyword(txt)
        inputs=inputs_show_user=f"Extract all image urls in this html page, pick the first 5 images and show them with markdown format: \n\n {page_return}"
        gpt_say = yield from request_gpt_model_in_new_thread_with_ui_alive(
            inputs=inputs, inputs_show_user=inputs_show_user,
            llm_kwargs=llm_kwargs, chatbot=chatbot, history=[], 
            sys_prompt="When you want to show an image, use markdown format. e.g. ![image_description](image_url). If there are no image url provided, answer 'no image url provided'"
        )
        chatbot[-1] = [chatbot[-1][0], gpt_say]
        yield from update_ui(chatbot=chatbot, history=history) # Refresh the page
        return



# ---------------------------------------------------------------------------------

def get_image_page_by_keyword(keyword):
    import requests
    from bs4 import BeautifulSoup
    response = requests.get(f'https://wallhaven.cc/search?q={keyword}', timeout=2)
    res = "image urls: \n"
    for image_element in BeautifulSoup(response.content, 'html.parser').findAll("img"):
        try:
            res += image_element["data-src"]
            res += "\n"
        except:
            pass
    return res
